import sys, os
from pathlib import Path
from typing import Any
import argparse

import numpy as np
import scipy as sp
import matplotlib as mpl
import matplotlib.pyplot as plt
from astropy.io import fits
import skimage as ski
import skimage.measure
import skimage.morphology


from aopp_deconv_tool import plot_helper
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
from aopp_deconv_tool.geometry.bounding_box import BoundingBox

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')



def select_labels_from_props(props, predicate = lambda prop: True):
	selected_labels = []
	for prop in props:
		if predicate(prop):
			selected_labels.append(prop.label)
	return selected_labels
		

def props_keep(props, labels_to_keep):
	props_to_keep = []
	for prop in props:
		if prop.label in labels_to_keep:
			props_to_keep.append(prop)
	return props_to_keep

def props_relabel(props, old_labels, new_labels):
	for prop in props:
		if prop.label in old_labels:
			prop.label = new_labels[old_labels.index(prop.label)]
	return props

def labels_keep(labels, labels_to_keep, background=0):
	for i in range(1, np.max(labels)+1):
		if i not in labels_to_keep:
			labels[labels==i] = background
	return labels

def labels_relabel(labels, old_labels, new_labels):
	for i, lbl in enumerate(old_labels):
		labels[labels==lbl] = new_labels[i]
	return labels

def labels_and_props_relabel(labels, props, old_labels, new_labels):
	for prop in props:
		lbl = prop.label
		if lbl in old_labels:
			new_lbl = new_labels[old_labels.index(lbl)]

			labels[labels==lbl] = new_lbl
			prop.label = new_lbl
	return(labels, props)



def get_source_regions(
		data : np.ndarray, 
		name : str = '',
		signal_noise_ratio : float = 5, 
		smallest_obj_area_px : int = 5,
		bbox_inflation_factor : float = 1.1,
		bbox_inflation_px : int = 30,
		morphology_opening : int = 3
	):
	"""
	Return regions, bounding boxes, and parameters used to find emitting regions in `data`
	"""
	_lgr.debug(f'{name=}')
	_lgr.debug(f'{signal_noise_ratio=}')
	_lgr.debug(f'{smallest_obj_area_px=}')
	_lgr.debug(f'{bbox_inflation_factor=}')
	_lgr.debug(f'{bbox_inflation_px=}')
	

	data_object_mask = data >= (np.median(data)*signal_noise_ratio)

	if morphology_opening > 0:
		data_object_mask = ski.morphology.opening(
			data_object_mask,
			ski.morphology.disk(morphology_opening, decomposition='sequence')
		)
	elif morphology_opening < 0:
		data_object_mask = ski.morphology.closing(
			data_object_mask,
			ski.morphology.disk(-1*morphology_opening, decomposition='sequence')
		)


	labels = ski.measure.label(data_object_mask)
	
	props = ski.measure.regionprops(labels)

	# only keep regions larger than `smallest_obj_area_px`
	large_region_labels = select_labels_from_props(props, lambda prop: prop.num_pixels >= smallest_obj_area_px)
	props = props_keep(props, large_region_labels)
	labels = labels_keep(labels, large_region_labels)

	props = sorted(props, key= lambda x: x.num_pixels, reverse=True)
	
	labels = labels_keep(labels, [p.label for p in props])

	labels, props = labels_and_props_relabel(labels, props, [p.label for p in props], tuple(range(1,len(props)+1)))
	
	bboxes = tuple(BoundingBox.from_min_max_tuple(prop.bbox).inflate(bbox_inflation_factor, bbox_inflation_px) for prop in props)

	return (
		labels, 
		bboxes,
		dict(
			name = name,
			signal_noise_ratio = signal_noise_ratio, 
			smallest_obj_area_px = smallest_obj_area_px,
			bbox_inflation_factor = bbox_inflation_factor,
			bbox_inflation_px = bbox_inflation_px,
			morphology_opening = morphology_opening,
		),
	)


def plot_source_regions(
		data : np.ndarray,
		labels,
		bboxes : tuple[BoundingBox,...],
		params : dict[str,Any],
		fig = None,
		ax = None,
		fig_kwargs=None,
	):
	noise_estimate = np.median(data[labels==0])
	nplots = len(bboxes)+2
	nr, nc = int(nplots // np.sqrt(nplots)), int(np.ceil(nplots/(nplots//np.sqrt(nplots))))
	
	f,ax = plot_helper.ensure_fig_and_ax(fig, ax, fig_kwargs={'figsize':(12,8)}, subplot_spec=(nr, nc))
	
	f.suptitle(f'Regions: S/N >= {params["signal_noise_ratio"]}; n_pixels >= {params["smallest_obj_area_px"]}')

	for a in ax:
		a.xaxis.set_visible(False)
		a.yaxis.set_visible(False)

	ax[0].set_title(f'{params["name"]}')
	ax[0].imshow(data)

	ax[1].set_title(f'Labels')
	ax[1].imshow(labels)

	for i, bbox in enumerate(bboxes):
		rect = mpl.patches.Rectangle(*bbox.to_mpl_rect(),edgecolor='r', facecolor='none', lw=1)
		text = f'region {i+1}'

		region_data = data[bbox.to_slices()]
		conservative_sig_noise = np.max(region_data)/np.median(region_data)
		region_sig_noise = np.max(region_data)/noise_estimate


		ax[0].add_patch(rect)
		ax[0].text(*bbox.mpl_min_corner, text, color='r', horizontalalignment='left', verticalalignment='top')
		ax[i+2].imshow(region_data)
		ax[i+2].set_title(f'{text}: S/N={region_sig_noise:0.2g} {"x".join(str(x) for x in bbox.extent)}')
	
	# Delete unused plot frames
	for i in range(len(bboxes)+2, len(ax)):
		ax[i].remove()

	return fig, ax

def run(
		fits_spec : aph.fits.specifier.FitsSpecifier,
		output_path_format : FPath,
		n_sources,
		signal_noise_ratio : float = 5, 
		smallest_obj_area_px : int = 5,
		bbox_inflation_factor : float = 1.1,
		bbox_inflation_px : int = 30,
		morphology_opening : int  = 3,
		plot : bool = False
	):
	
	with fits.open(Path(fits_spec.path)) as data_hdul:
		
		
		_lgr.debug(f'{fits_spec.path=} {fits_spec.ext=} {fits_spec.slices=} {fits_spec.axes=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
		data = data_hdu.data[fits_spec.slices]
		hdr = data_hdu.header
		original_data_type=data_hdu.data.dtype
		axes = fits_spec.axes['CELESTIAL']
		
		assert all([s==1 for i, s in enumerate(data.shape) if i not in axes]), f"For extracting sources, we can only operate on a single 2D image not one of shape {data.shape}."
		
		if data.ndim > 2:
			squeeze_select = tuple(0 if i not in axes else slice(None) for i in range(data.ndim))
			undo_squeeze = tuple(None if i not in axes else slice(None) for i in range(data.ndim))
		else:
			squeeze_select = None
			undo_squeeze = None
		
		if squeeze_select is not None:
			data = data[*squeeze_select]
		
		
		labels, bboxes, params = get_source_regions(
			data,
			Path(fits_spec.path).name,
			signal_noise_ratio,
			smallest_obj_area_px,
			bbox_inflation_factor,
			bbox_inflation_px,
			morphology_opening,
		)
		
		if n_sources is not None:
			labels[labels > n_sources] = 0
			bboxes = bboxes[:n_sources]
		else:
			n_sources = len(bboxes)
		
		
		if plot:
			fig, ax = plot_source_regions(data, labels, bboxes, params)
			plot_helper.output(show=True, figure=fig)
		
		
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'n_sources' : n_sources,
			'signal_noise_ratio' : signal_noise_ratio,
			'smallest_obj_area_px' : smallest_obj_area_px,
			'bbox_inflation_factor' : bbox_inflation_factor,
			'bbox_inflation_px' : bbox_inflation_px,
			'morphology_opening' : morphology_opening,
		}
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='extract_sources',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
	
	for i in range(0, n_sources):
		hdus = []
	
		output_data = data[bboxes[i].to_slices()].astype(original_data_type)
		if undo_squeeze:
			output_data = output_data[*undo_squeeze]
	
		hdus.append(fits.PrimaryHDU(
			header = hdr,
			data = output_data
		))
		
		
		# Want source 0 to be background, but do not save source 0, so start from source 1
		output_path = output_path_format.with_fields(i=i+1)
		
		hdul_output = fits.HDUList(hdus)
		hdul_output.writeto(output_path, overwrite=True)
		_lgr.info(f'Written source {i+1} to {output_path.relative_to(Path(),walk_up=True)}')
	
	


def parse_args(argv):
	parser = argparse.ArgumentParser(
		description=__doc__,
	)
	
	DEFAULT_OUTPUT_TAG = '_source'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter,
		epilog=FITS_SPECIFIER_HELP
	)
	
	parser.add_argument(
		'fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon . See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	
	parser.add_argument(
		'-o', 
		'--output_path_format', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}_{i}{suffix}',
		help = '\n    '.join((
			f'Output fits file path format, supports keyword substitution using parts of `fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".", including the ".")',
			'{i} : The source number from brightest to dimmest.',
			'\b'
		))
	)
	
	parser.add_argument(
		'-n',
		'--n_sources',
		type = int,
		default = None,
		help='Maximum number of sources that will be foundnd and made into their own file. Will return largest sources first.'
	)
	
	parser.add_argument(
		'--signal_noise_ratio',
		type=float,
		default = 5, 
		help='Minimum signal/noise for a source to be found'
	)
	parser.add_argument(
		'--smallest_obj_area_px',
		type=int,
		default = 5,
		help='Smallest number of pixels above S/N for a source to be found'
	)
	parser.add_argument(
		'--bbox_inflation_factor',
		type=float,
		default=1.1,
		help='Minimum factor by which to inflate the bounding box of a found source'
	)
	parser.add_argument(
		'--bbox_inflation_px',
		type= int,
		default = 30,
		help='Minimum number of pixels by which to inflate the bounding box of a found source.'
	)
	parser.add_argument(
		'--morphology_opening',
		type=int,
		default=3,
		help='Apply morphology opening to S/N ratio map of observation. Large value groups nearby sources together.'
	)
	
	
	parser.add_argument(
		'--plot',
		action='store_true',
		help='If present, will plot the regions found before writing them to disk.'
	)
	
	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, DESIRED_FITS_AXES)
	
	other_file_path = Path(args.fits_spec.path)
	args.output_path_format = FPath(
		args.output_path_format.with_fields(
			parent=other_file_path.parent, 
			stem=other_file_path.stem, 
			suffix=other_file_path.suffix,
			tag=DEFAULT_OUTPUT_TAG,
			i='{i}'
		)
	)
	
	for k,v in vars(args).items():
		_lgr.debug(f'{k} = {v}')
	
	return args



def go(
		fits_spec,
		output_path_format=None,
		n_sources = None,
		signal_noise_ratio = None, 
		smallest_obj_area_px = None, 
		bbox_inflation_factor = None, 
		bbox_inflation_px = None, 
		morphology_opening = None,
		plot_FLAG=None,
		help_FLAG=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""	
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args = parse_args(argv)
	
	run(
		args.fits_spec,
		output_path_format=args.output_path_format,
		n_sources = args.n_sources,
		signal_noise_ratio = args.signal_noise_ratio, 
		smallest_obj_area_px = args.smallest_obj_area_px, 
		bbox_inflation_factor = args.bbox_inflation_factor, 
		bbox_inflation_px = args.bbox_inflation_px, 
		morphology_opening = args.morphology_opening,
		plot = args.plot
	)

if __name__=='__main__':
	exec_with_args(sys.argv[1:])







from __future__ import annotations
from typing import TypeVar, TypeVarTuple, Type, Generic, GenericAlias
import numpy as np

from aopp_deconv_tool.typedef import NumVar, ShapeVar


IntVar = TypeVar('IntVar', bound=int)
T = TypeVar('T')
Ts = TypeVarTuple('Ts')
S = GenericAlias(tuple, (IntVar,))
N = TypeVar('N',bound=int)



class BoundingBox(Generic[N,T]):
	"""
	Class to handle the concept of a bounding box in some coordinate system
	"""
	def __init__(self, 
			corner_min : np.ndarray[S[N], T],
			corner_max : np.ndarray[S[N], T],
			dtype : Type | None = None
		):
		self.coords = np.array([corner_min, corner_max],dtype=dtype if dtype is not None else corner_min.dtype)
	
	@classmethod
	def from_min_max_tuple(cls, 
			min_max_tuple : tuple[*Ts]
		) -> BoundingBox[N,T]:
		n = int(len(min_max_tuple)//2)
		return	cls(np.array(min_max_tuple[:n]), np.array(min_max_tuple[n:]))

	@classmethod
	def from_mpl_rect(cls, 
			mpl_rect : tuple[tuple[T,T],T,T]
		) -> BoundingBox[N,T]:
		n = 2
		min_corner = np.array(mpl_rect[0])
		max_corner = min_corner + np.array(mpl_rect[1:])
		return cls(min_corner, max_corner)
	
	@classmethod
	def from_slices(cls, 
			slices : tuple[slice,...]
		) -> BoundingBox[N,T]:
		min_corner = np.array([s.start if s.start is not None else 0 for s in slices])
		max_corner = np.array([s.stop if s.stop is not None else -1 for s in slices])
		return cls(min_corner, max_corner)
		
	def to_mpl_rect(self) -> tuple[tuple[T,T],T,T]:
		return tuple(self.coords[0][::-1]), *tuple((self.coords[1]-self.coords[0])[::-1])
	
	def to_slices(self) -> tuple[slice,...]:
		return tuple(slice(*x) for x in self.coords.T)
	
	def inflate(self, factor, n) -> BoundingBox:
		centre = (self.coords[1] + self.coords[0])/2
		half_extent = self.extent/2
		self.coords = np.array(((centre - (factor*half_extent+n)),(centre + (factor*half_extent+n))), dtype=self.coords.dtype)
		return self

	def __repr__(self) -> str:
		return repr(self.coords)
	
	@property
	def min_corner(self) -> np.ndarray[S[N],T]:
		return self.coords[0]

	@property
	def max_corner(self) -> np.ndarray[S[N],T]:
		return self.coords[1]
	
	@property
	def mpl_min_corner(self) -> np.ndarray[S[N],T]:
		return self.coords[0][::-1]

	@property
	def mpl_max_corner(self) -> np.ndarray[S[N],T]:
		return self.coords[1][::-1]
	
	@property
	def extent(self) -> np.ndarray[S[N],T]:
		return self.coords[1] - self.coords[0]





from typing import Annotated

import numpy as np

from aopp_deconv_tool.optics.function import PhasePowerSpectralDensity

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


"""
	objective                                    image plane
	-----------------------------------------------|
		|                                          |
		|                           .............. X         point on image plane, priciple ray to this point is ".../" line
		|             ............./   \           |
		|............/           theta  |          |
......../ - - - - - - - - - - - - - - - - - - - - -|         optical axis
		|                                          |
		|                                          |
		|                                          |
		|                                          |
	-----------------------------------------------|

theta - Angle between optical axis and principle ray to point X on image plane

"""

def phase_psd_von_karman_turbulence(
		f_axes : Annotated[np.ndarray, 'ModelInputData'], # f_axes, f is theta/wavelength
		wavelength : Annotated[float, 'ModelInputData'],
		r0 : Annotated[float, 'ModelParameter'],
		turbulence_ndim : Annotated[float,'ModelParameter'],
		L0 : Annotated[float,'ModelParameter']
	):
	"""
	Von Karman turbulence is the same as Kolmogorov, but with the extra "L0"
	term. L0 -> infinity gives Kolmogorov turbulence.
	"""
	wav_factor = wavelength/5E-7
	L0 = L0*wav_factor
	#_lgr.debug(f'{f_axes=} {wavelength=} {wav_factor=} {r0=} {turbulence_ndim=} {L0=}')
	
	f_mesh = np.array(np.meshgrid(*f_axes[::-1]))
	f_sq = np.sum(f_mesh**2, axis=0)
	r0_pow = -5/3
	f_pow = -(2+turbulence_ndim*3)/6
	factor = 0.000023*10**(turbulence_ndim)
	r0_wavelength_corrected = r0*wav_factor**(-1/r0_pow)
	
	psd = factor*(r0_wavelength_corrected)**(r0_pow)*(1/L0**2 + f_sq)**(f_pow)
	#centre_idx = tuple(s//2 for s in psd.shape)
	#psd[centre_idx] = 0 # stop infinity at f==0
	return PhasePowerSpectralDensity(psd, f_axes)




"""
Graphical front end to SSA filtering of an image, should be usable for a cube and a 2d image
"""

import sys, os
from pathlib import Path
from typing import Callable, Any

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec
import matplotlib.widgets
import matplotlib.ticker
import matplotlib.transforms
import matplotlib.patches
import matplotlib.lines

import PIL.Image

from aopp_deconv_tool.py_ssa import SSA

from aopp_deconv_tool.plot_helper import fig_draw_bbox_of_artist

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


default_scalar_formatter = mpl.ticker.ScalarFormatter().format_data


class CallbackMixin:
	def __init__(self, *args : tuple[str]):
		self.callbacks = {}
		for csn in args:
			self.callbacks[csn] = []
		
	def get_callback_sets(self):
		return self.callbacks.keys()
	
	def set_callback_sets(self, callback_set_names):
		for csn in callback_set_names:
			if csn not in self.callbacks:
				self.callbacks[csn] = []
	
	def remove_callback_sets(self, callback_set_names):
		for csn in callback_set_names:
			del self.callbacks[csn]
	
	def attach_callback(self, callback_set, acallable : Callable[[Any,...],bool]) -> int:
		try:
			i = self.callbacks[callback_set].index(None)
			self.callbacks[callback_set][i] = acallable
			return i
		except ValueError:
			i = len(self.callbacks[callback_set])
			self.callbacks[callback_set].append(acallable)
		return i
	
	def disconnect_callback_idx(self, callback_set, idx):
		self.callbacks[callback_set][idx] = None
	
	def call_callbacks(self, callback_set, *args, **kwargs):
		for acallback in self.callbacks.get(callback_set,[]):
			if acallback is not None: 
				acallback(*args, **kwargs)


class BasePlot:
	def __init__(self, ax, **kwargs):
		self.ax = ax
		self.data = None
		self.kwargs = kwargs
		# self.handle holds the things that are shown on the axes.
		self.handle = None
	
	def remove(self):
		if self.handle is not None:
			# if the handle is a list, then remove each element individually
			# otherwise, remove the handle itself
			if type(self.handle) in (list, tuple):
				for h in self.handle:
					h.remove()
			else:
				self.handle.remove()
		self.handle = None
	
	def set_title(self, title : None | str):
		self.ax.set_title(title)
	

class ImagePlot(BasePlot, CallbackMixin):
	def __init__(self, ax, **kwargs):
		CallbackMixin.__init__(self,'on_set_clim','on_set_data')
		BasePlot.__init__(self, ax, **kwargs)
	
	def set_clim(self, min=None, max=None):
		if min is None or max is None:
			image_data = self.handle.get_array()
			if min is None: min = np.nanmin(image_data)
			if max is None: max = np.nanmax(image_data)
		self.handle.set(clim=(min,max))
		self.call_callbacks('on_set_clim', min, max)
		
	
	def set_axes_limits(self, xlim=(None,None), ylim=(None,None)):
		self.ax.set_xlim(*xlim)
		self.ax.set_ylim(*ylim)
	
	
	def sanitise_data(self, new_data):
		if type(new_data) is not np.ndarray:
			raise RuntimeError(f'{self.__class__.__name__} accepts 2 dimensional numpy arrays')
		if new_data.ndim != 2:
			raise RuntimeError(f'{self.__class__.__name__} accepts 2 dimensional numpy arrays as input data, was passed data with {new_data.ndim} dimensions')
		return new_data
	
	def is_new_data_compatible(self, sanitised_data):
		if self.data is None:
			return False
		return all(s1 == s2 for s1,s2 in zip(self.data.shape, sanitised_data.shape))
	
	def remove(self):
		if self.handle is not None:
			self.handle.remove()
		self.handle = None
	
	def set_data(self, new_data):
		sanitised_new_data = self.sanitise_data(new_data)
		if not self.is_new_data_compatible(sanitised_new_data):
			self.remove()
			self.data = sanitised_new_data
			self.create()
		else:
			self.data = sanitised_new_data
			self.handle.set_data(self.data)
		self.set_axes_limits()
		self.set_clim()
		self.call_callbacks('on_set_data', self)
	
	
	def create(self):
		self.handle = self.ax.imshow(self.data, **self.kwargs)


class LinePlot(BasePlot, CallbackMixin):
	def __init__(self, ax, **kwargs):
		CallbackMixin.__init__(self,'on_set_data')
		BasePlot.__init__(self, ax, **kwargs)

	def is_new_data_compatible(self, sanitised_new_data):
		if (self.data is None
			or len(sanitised_new_data) > len(self.data)
		):
			return False
		return True

	

	def sanitise_data(self, new_data):
		# want to put data into a format where we have
		# three dimensions: "dataset", "x or y data", "values"
		# I.e., the shape is always (n_datasets, 2, n_points)
		ndim = 0
		_lgr.debug(f'{type(new_data)=}')
		if type(new_data) is np.ndarray:
			ndim = new_data.ndim
		else:
			# how many levels can we index
			test_value = new_data
			while hasattr(new_data, "__getitem__"):
				ndim += 1
				test_value = new_data[0]
		
		match (ndim):
			case 1:
				return ((np.arange(len(new_data)),new_data),)
			case 2:
				return (new_data,)
			case 3:
				return new_data
			case _:
				raise ValueError(f'{self.__class__.__name__} can only accept data with up to 3 dimensions of shape (n_datasets, 2, n_points). Passed data has {ndim} dimensions.')
			
		
	def set_axes_limits(self, 
			xlim : tuple[float | None, float | None] = (None,None), 
			ylim  : tuple[float | None, float | None] = (None,None)
		):
		self.ax.set_xlim(*xlim)
		self.ax.set_ylim(*ylim)
		


	def set_data(self, new_data):
		sanitised_new_data = self.sanitise_data(new_data)
		if not self.is_new_data_compatible(sanitised_new_data):
			self.remove()
			self.data = sanitised_new_data
			self.create()
		else:
			for i, line in enumerate(self.handle):
				if i < len(sanitised_new_data):
					line.set_data(sanitised_new_data[i])
				else:
					line.remove()
		self.set_axes_limits()
		self.call_callbacks('on_set_data', self)
	
	def create(self):
		self.handle = []
		for dataset in self.data:
			self.handle.extend(self.ax.plot(dataset[0], dataset[1], **self.kwargs))
				

class WidgetBase:
	def __init__(self, fig, ax_rect):
		self.ax = fig.add_axes(ax_rect)
		self.widget = None

	def set_active(self, active : bool):
		super(mpl.widgets.AxesWidget, self.widget).set_active(active)
	
	def get_active(self):
		return super(mpl.widgets.AxesWidget, self.widget).active
	
	def set_value(self, value):
		self.widget.set_val(value)
	
	def get_value(self):
		return self.widget.val
	
	def disconnect(self, cid):
		return self.widget.disconnect(cid)
	
	def remove(self):
		self.ax.remove()



class Slider(WidgetBase):
	def __init__(self, fig, ax_rect, min=0, max=0, step=None, label='slider', orientation='vertical', **kwargs):
		super().__init__(fig, ax_rect)
		
		self.widget = mpl.widgets.Slider(self.ax, valmin=min, valmax=max, valinit=min, valstep=step, label=label, orientation=orientation)
	
		self.set_labels_for_orientation(fig, orientation)
	
	def set_active(self, active : bool):
		if not active:
			self.set_limits(0,0)
		super().set_active(active)
	
	def set_labels_for_orientation(self, fig, orientation):
		renderer = fig.canvas.get_renderer()
		if orientation == 'vertical':
			h_pos = 0.2
			self.widget.label.set_ha('right')
			self.widget.label.set_va('bottom')
			self.widget.label.set_rotation('vertical')
			self.widget.label.set_position([h_pos,0.0])
			label_bb = self.ax.transAxes.inverted().transform_bbox(self.widget.label.get_window_extent(renderer))
			
			self.widget.valtext.set_ha('right')
			self.widget.valtext.set_va('bottom')
			self.widget.valtext.set_rotation('vertical')
			self.widget.valtext.set_position([h_pos,label_bb.height+0.01])
		
	
	def set_limits(self, min = None, max = None):
		if min is None: min = self.widget.valmin
		if max is None: max = self.widget.valmax
		assert min <= max, f'{self.__class__.__name__}.set_limits(...), min must be <= max, currently {min=} {max=}'
		self.widget.valmin = min
		self.widget.valmax = max
		
		value = self.get_value()
		if value < min:
			self.set_value(min)
		elif value > max:
			self.set_value(max)
		
		match self.widget.orientation:
			case 'vertical':
				self.ax.set_ylim(min,max)
			case 'horizontal':
				self.ax.set_xlim(min,max)
			case _:
				raise RuntimeError(f'Slider orientation "{self.widget.orientation}" is unknown, allowed values "vertical" or "horizontal"')
	
	def on_changed(self, acallable : Callable[[tuple[float,float]],bool]):
		def callback(x):
			if acallable(x):
				self.ax.figure.canvas.draw()
				
		return self.widget.on_changed(callback)


class CheckButtons(WidgetBase):
	def __init__(self, fig, ax_rect, labels=['checkbutton 1', 'checkbutton 2'],**kwargs):
		super().__init__(fig, ax_rect)
		self.labels = labels
		self.widget = mpl.widgets.CheckButtons(self.ax, labels=self.labels, **kwargs)
	
	def on_clicked(self, dict_of_callables : dict[str,Callable[[bool],None]]):
		def dispatch_check_events(label_toggled):
			callback = dict_of_callables.get(label_toggled,None)
			if callback is None: return

			state = self.get_value()[self.labels.index(label_toggled)]
			if callback(state):
				self.ax.figure.canvas.draw()
			
		return self.widget.on_clicked(dispatch_check_events)
	
	def get_value(self):
		return self.widget.get_status()
	
	def set_value(self, values):
		current_values = self.get_value()
		for i, (v, c) in enumerate(zip(values, current_values)):
			if v != c:
				self.widget.set_active(i)

class TextBox(WidgetBase):
	def __init__(self, fig, ax_rect, label='textbox', initial='', **kwargs):
		super().__init__(fig, ax_rect)
		self.widget = mpl.widgets.TextBox(self.ax, label=label, initial=initial, **kwargs)
		
	def get_value(self):
		return self.widget.text#.get_text()
	
	def on_submit(self, acallable : Callable[[str],bool]):
		def callback(text):
			if acallable(text.get_text()):
				self.ax.figure.canvas.draw()
		return self.widget.on_submit(callback)


class Range(Slider):
	def __init__(self, fig, ax_rect, label='range', min=-1E300, max=1E300, orientation='vertical', step=None, **kwargs):
		self.ax = fig.add_axes(ax_rect)
		self.widget = mpl.widgets.RangeSlider(self.ax, label, valmin=min, valmax=max, orientation=orientation, valstep=step, **kwargs)
		
		self.set_labels_for_orientation(fig, orientation)
			
	
	def set_limits(self, min = None, max = None, sticky_values=[False,False]):
		if min is None: min = self.widget.valmin
		if max is None: max = self.widget.valmax
		assert min < max, f'{self.__class__.__name__}.set_limits(...), min must be <= max, currently {min=} {max=}'
		self.widget.valmin = min
		self.widget.valmax = max
		minmax = (min,max)
		
		value = list(self.get_value())
		for i in range(len(value)):
			if sticky_values[i]:
				if value[i] < min:
					value[i] = min
				elif value[i] > max:
					value[i] = max
			else:
				value[i] = minmax[i]
		
		self.set_value(value)
		
		match self.widget.orientation:
			case 'vertical':
				self.ax.set_ylim(min,max)
			case 'horizontal':
				self.ax.set_xlim(min,max)
			case _:
				raise RuntimeError(f'Slider orientation "{self.widget.orientation}" is unknown, allowed values "vertical" or "horizontal"')


class RadioButtons(WidgetBase):
	def __init__(self, fig, ax_rect, labels, title='', **kwargs):
		super().__init__(fig, ax_rect)
		self.labels = labels
		self.widget = mpl.widgets.RadioButtons(self.ax, self.labels, **kwargs)
		self.title_text = self.ax.text(0.1,1.05,title, transform=self.ax.transAxes, va='top', ha='left')
		self.ax.axis('off')
		
	def set_value(self, idx):
		self.widget.set_active(idx)

	def get_value(self):
		return self.widget.active
	
	def on_clicked(self, callback : Callable[[int],bool]):
		def dispatch_click_event(selected_label):
			if callback(selected_label):
				self.ax.figure.canvas.draw()
		self.widget.on_clicked(dispatch_click_event)


class Button(WidgetBase):
	def __init__(self, fig, ax_rect, label, **kwargs):
		super().__init__(fig, ax_rect)
		self.widget = mpl.widgets.Button(self.ax, label, **kwargs)
	
	def on_clicked(self, callback : Callable[[],bool]):
		def dispatch_click_event(selected_label):
			if callback(selected_label):
				self.ax.figure.canvas.draw()
		self.widget.on_clicked(dispatch_click_event)




class BasePanel:
	def __init__(self, 
			parent_figure : None | mpl.figure.Figure | mpl.figure.SubFigure = None, 
			pf_gridspec : None | mpl.gridspec.GridSpec = None, 
			gridspec_index : int = 0, 
			window_title : None | str = None
		):
		# Information about how to draw the panel in the containing figure
		self.parent_figure = plt.figure(figsize=(12,8)) if parent_figure is None else parent_figure
		self.pf_gridspec = self.parent_figure.add_gridspec(1,1,left=0,right=0,top=1,bottom=1,wspace=0,hspace=0) if pf_gridspec is None else pf_gridspec
		self.window_title=window_title
		
		self.set_window_title()
		
		# Figure we are going to use to draw the panel
		self.figure = self.parent_figure.add_subfigure(self.pf_gridspec[gridspec_index], frameon=False)
		
		self.create_main_axes()
		self.create_main_handles()
	
	def create_main_axes(self):
		# Get the axes for the plot
		# (left, bottom, width, height) in fraction of figure area
		self.main_axes_rect = (0.2, 0.2, 0.7, 0.7)
		self.main_axes = self.figure.add_axes(self.main_axes_rect, projection=None, polar=False)
	
	def create_main_handles(self):
		self.main_data_handle = None
		self.main_plot_handle = None

	def set_window_title(self):
		if self.window_title is not None:
			self.figure.canvas.manager.set_window_title(self.window_title)
	
	def set_data(self, new_data):
		NotImplemented
	
	def show(self, data : None | Any = None, title : None | str = None):
		plt.figure(self.parent_figure)
		
		self.set_data(data)
		self.main_axes.set_title(title)
		
		plt.show()
	


class ImageViewer(CallbackMixin):
	def __init__(self, parent_figure=None, pf_gridspec = None, gridspec_index = 0, window_title='Image Viewer'):
		
		self.parent_figure = plt.figure(figsize=(12,8)) if parent_figure is None else parent_figure
		self.pf_gridspec = self.parent_figure.add_gridspec(1,1,left=0,right=0,top=1,bottom=1,wspace=0,hspace=0) if pf_gridspec is None else pf_gridspec
		self.figure = self.parent_figure.add_subfigure(self.pf_gridspec[gridspec_index], frameon=False)
		
		self.window_title=window_title
		self.set_window_title()
		
		self.main_axes_rect = (0.2, 0.2, 0.7, 0.7) # (left, bottom, width, height) in fraction of figure area
		self.main_axes = self.figure.add_axes(self.main_axes_rect, projection=None, polar=False)
		self.main_axes_image_data = None
		self.main_axes_im = ImagePlot(self.main_axes)
		
		self.image_plane_slider = Slider(self.figure, (0.02, 0.2, 0.02, 0.7), step=1, label='image plane', active=False)
		self.image_plane_slider.on_changed(self.set_image_plane)
		
		self.main_axes_visibility_controls = CheckButtons(self.figure, (0.08,0.8,0.06,0.1), ['xaxis','yaxis'], actives=[True,True])
		self.main_axes_visibility_controls.on_clicked(
			{	'xaxis' : lambda x: (self.main_axes.xaxis.set_visible(x),True)[1],
				'yaxis': lambda x: (self.main_axes.yaxis.set_visible(x), True)[1]
			}
		)
		self.main_axes_visibility_controls.set_value([False,False])
		
		self.image_clim_slider = Range(self.figure, (0.05, 0.2, 0.02, 0.7), label='clims')
		self.image_clim_slider.on_changed(lambda x: self.main_axes_im.set_clim(*x))
		self.image_clim_slider_modes = (
			'displayed data limits',
			'all data limits'
		)
		self.image_clim_slider_mode_callback_id = None
		
		self.image_clim_slider_mode_selector = RadioButtons(self.figure, (0.06,0.65,0.2,0.1), self.image_clim_slider_modes, title='clim slider mode')
		self.set_clim_slider_mode(self.image_clim_slider_modes[0])
		self.image_clim_slider_mode_selector.on_clicked(self.set_clim_slider_mode)
	
	def set_window_title(self):
		if self.window_title is not None:
			self.figure.canvas.manager.set_window_title(self.window_title)
	
	def set_clim_slider_mode(self, mode):
		if self.image_clim_slider_mode_callback_id is not None:
			self.main_axes_im.disconnect_callback_idx('on_set_data', self.image_clim_slider_mode_callback_id)
		
		match self.image_clim_slider_modes.index(mode):
			case 0:
				# Have clim slider always be min-max of new data plane
				self.image_clim_slider_mode_callback_id = self.main_axes_im.attach_callback(
					'on_set_data', 
					lambda other: self.image_clim_slider.set_limits(np.nanmin(other.data), np.nanmax(other.data))
				)
				if self.main_axes_image_data is not None:
					self.image_clim_slider.set_limits(
						np.nanmin(self.get_displayed_data()), 
						np.nanmax(self.get_displayed_data())
					)
			
			case 1:
				# Have clim slider holder the full range of data, and be min-max of new data plane when selected
				self.image_clim_slider_mode_callback_id = self.main_axes_im.attach_callback(
					'on_set_data', 
					lambda other: self.image_clim_slider.set_value((np.nanmin(other.data), np.nanmax(other.data)))
				)
				if self.main_axes_image_data is not None:
					self.image_clim_slider.set_limits(
						np.nanmin(self.main_axes_image_data), 
						np.nanmax(self.main_axes_image_data)
					)
			
			case _:
				ValueError(f'Unrecognised slider mode "{mode}", must be one of {self.image_clim_slider_modes}')
			
		if self.main_axes_image_data is not None:
			self.image_clim_slider.set_value(
				(np.nanmin(self.get_displayed_data()), np.nanmax(self.get_displayed_data()))
			)
		return True
		
	
	def get_displayed_data(self):
		return self.main_axes_image_data[self.image_plane_slider.get_value()] if self.image_plane_slider.get_active() else self.main_axes_image_data
	
	def get_data(self):
		return self.main_axes_image_data
	
	def set_data(self, data : np.ndarray = None):
		self.main_axes_image_data = data
		if self.main_axes_image_data is None: return
		
		
		if self.main_axes_image_data.ndim == 2:
			self.image_plane_slider.set_active(False)
		
			
		elif self.main_axes_image_data.ndim == 3:
			_lgr.debug(f'{self.main_axes_image_data.shape[0]=}')
			self.image_plane_slider.set_active(True)
			self.image_plane_slider.set_limits(0,self.main_axes_image_data.shape[0]-1)
		
		self.main_axes_im.set_data(self.get_displayed_data())
			
	def set_image_plane(self, x):
		self.main_axes_im.set_data(self.main_axes_image_data[int(x)])
		return True

	def show(self, data=None, title=None):
		plt.figure(self.parent_figure)
		
		self.set_data(data)
		if title is not None:
			self.main_axes.set_title(title)
		
		plt.show()



class ResidualViewer(ImageViewer):
	def __init__(self, parent_figure=None, pf_gridspec = None, gridspec_index = 0, window_title='Image Viewer'):
		super().__init__(parent_figure, pf_gridspec, gridspec_index, window_title)
		self.frozen_clims = (0,1)
	
	
	def set_clim_slider_mode(self, mode):
		if self.image_clim_slider_mode_callback_id is not None:
			self.main_axes_im.disconnect_callback_idx('on_set_data', self.image_clim_slider_mode_callback_id)
		
		match self.image_clim_slider_modes.index(mode):
			case 0:
				# Have clim slider always be min-max of new data plane
				self.image_clim_slider_mode_callback_id = self.main_axes_im.attach_callback(
					'on_set_data', 
					lambda other: self.image_clim_slider.set_limits(np.nanmin(other.data), np.nanmax(other.data))
				)
				if self.main_axes_image_data is not None:
					self.image_clim_slider.set_limits(
						np.nanmin(self.get_displayed_data()), 
						np.nanmax(self.get_displayed_data())
					)
			
			case 1:
				# Have clim slider hold the frozen limits but the image values value
				self.image_clim_slider_mode_callback_id = self.main_axes_im.attach_callback(
					'on_set_data', 
					lambda other: self.image_clim_slider.set_value((np.nanmin(other.data), np.nanmax(other.data)))
				)
				if self.main_axes_image_data is not None:
					self.image_clim_slider.set_limits(*self.frozen_clims)
			
			case _:
				ValueError(f'Unrecognised slider mode "{mode}", must be one of {self.image_clim_slider_modes}')
			
		if self.main_axes_image_data is not None:
			self.image_clim_slider.set_value(
				(np.nanmin(self.get_displayed_data()), np.nanmax(self.get_displayed_data()))
			)
		return True
	

class ImageAggregator(ImageViewer, CallbackMixin):
	def __init__(self, parent_figure=None, pf_gridspec = None, gridspec_index = 0, window_title='Image Aggregator'):
		CallbackMixin.__init__(self, 'on_set_image_agg_planes')
		ImageViewer.__init__(self, parent_figure, pf_gridspec , gridspec_index , window_title)
		
		self.image_plane_slider.remove()
		self.image_plane_slider = Range(self.figure, (0.02, 0.2, 0.02, 0.7), step=1, label='aggregate image planes')
		self.image_plane_slider.on_changed(self.set_image_agg_planes)
		
		
	
	def get_displayed_data(self):
		x = self.image_plane_slider.get_value()
		return np.nansum(self.main_axes_image_data[int(x[0]):int(x[1])], axis=0) if self.image_plane_slider.get_active() else self.main_axes_image_data
	
	
	def set_image_agg_planes(self, x):
		self.main_axes_im.set_data(np.nansum(self.main_axes_image_data[int(x[0]):int(x[1])], axis=0))
		self.call_callbacks('on_set_image_agg_planes', self)
		return True
	
	def set_data(self, data : np.ndarray = None):
		super().set_data(data)
		_lgr.debug(f'{data.shape=}')
		self.image_plane_slider.set_limits(0,data.shape[0])
		self.image_plane_slider.set_value((0,data.shape[0]))
		

class LineRegionPanel(BasePanel, CallbackMixin):
	class RemoveRegionException(Exception):
		pass

	def __init__(self, parent_figure=None, pf_gridspec = None, gridspec_index = 0, window_title='Line Region', **kwargs):
		CallbackMixin.__init__(self, 'on_region_changed')
		super().__init__(parent_figure, pf_gridspec, gridspec_index, window_title)
		
		self.main_plot_handle = LinePlot(self.main_axes, **kwargs)
		self.main_data_handle = None
		
		self.shown_span_hdls = None
		self.span_selector_wdgt = mpl.widgets.SpanSelector(
			self.main_axes,
			direction='horizontal',
			onselect = self.region_changed
		)
		self.set_region_limits()
	
	def set_data(self, new_data):
		self.main_data_handle = new_data
		self.main_plot_handle.set_data(self.main_data_handle)
	
	def set_region_limits(self, 
			min_value : None | float | int = None, 
			max_value : None | float | int = None, 
			snap_step_value : None | float | int = None, 
			min_size : None | float | int = None, 
			max_size : None | float | int = None,
			snap_action = 'round', 
			below_min_action='set', 
			above_max_action='set',
			region_too_small_action = 'remove_region',
			region_too_large_action = 'remove_region'
		):
		self.r_min_value = min_value
		self.r_max_value = max_value
		self.r_snap_step_value = snap_step_value
		self.r_min_size = min_size
		self.r_max_size = max_size
		self.r_snap_action = snap_action
		self.r_below_min_action = below_min_action
		self.r_above_max_action = above_max_action
		self.r_region_too_small_action = region_too_small_action
		self.r_region_too_large_action = region_too_large_action
	
	
	
	
	def apply_region_action(self, value: int | float, limit : int | float, action : str) -> int | float | None:
		match action:
			case 'round':
				remainder = value % limit
				return value - remainder if (remainder < 1/2*limit) else value - remainder + limit
			case 'floor':
				return value - value % limit
			case 'ceil':
				return (value + limit) % limit
			case 'set':
				return limit
			case 'remove_region':
				raise LineRegionPanel.RemoveRegionException
	
	def apply_region_limits(self, r_min, r_max):
		# Apply min/max values
		if self.r_min_value is not None:
			if r_min < self.r_min_value:
				r_min = self.apply_region_action(r_min, self.r_min_value, self.r_below_min_action)
			if r_max < self.r_min_value:
				r_max = self.apply_region_action(r_max, self.r_min_value, self.r_below_min_action)
		
		if self.r_max_value is not None:
			if r_min > self.r_max_value:
				r_min = self.apply_region_action(r_min, self.r_max_value, self.r_below_max_action)
			if r_max > self.r_max_value:
				r_max = self.apply_region_action(r_max, self.r_max_value, self.r_below_max_action)
		
		# Apply snap to step
		if self.r_snap_step_value is not None:
			r_min = self.apply_region_action(r_min, self.r_snap_step_value, self.r_snap_action)
			r_max = self.apply_region_action(r_max, self.r_snap_step_value, self.r_snap_action)
		
		if self.r_min_size is not None and (r_max - r_min) < self.r_min_size:
			region_size = self.apply_region_action((r_max - r_min), self.r_min_size, self.r_region_too_small_action)
			r_max = r_min + region_size
		
		if self.r_max_size is not None and (r_max - r_min) > self.r_max_size:
			region_size = self.apply_region_action((r_max - r_min), self.r_max_size, self.r_region_too_large_action)
			r_max = r_min + region_size
		return r_min, r_max
		
		
	
	def region_changed(self, r_min : float, r_max : float) -> bool:
		try:
			r_min, r_max = self.apply_region_limits(r_min, r_max)
		except LineRegionPanel.RemoveRegionException:
			self.remove_region()
			self.call_callbacks('on_region_changed', None, None)
			return
			
		self.set_region(r_min, r_max)
		self.call_callbacks('on_region_changed', r_min, r_max)
	
	def remove_region(self):
		if self.shown_span_hdls is not None:
			for h in self.shown_span_hdls:
				h.remove()
			self.shown_span_hdls = None
	
	def set_region(self, r_min, r_max):
		self.remove_region()
		self.shown_span_hdls = [
			self.main_axes.axvspan(r_min, r_max, alpha=0.3, color='tab:red'),
			self.main_axes.axvline(r_min, alpha=0.5, color='tab:red', ls='-'),
			self.main_axes.axvline(r_max, alpha=0.5, color='tab:red', ls='--')
		]
	

class SSAViewer:
	"""
	Graphical front-end to the singular spectrum analysis routines in the `aopp_deconv_tool.py_ssa` module.
	Shows the input data, the eigenvalues of the SSA components, an aggregate of SSA components, and a residual
	between the aggregate and the input data.
	
	Accepts numpy arrays to the `SSAViewer.show(image:np.ndarray)` method. Ouputs numpy arrays (`.npy` file)
	or TIFF (`.tif` file) images.
	"""
	def __init__(self, 
			parent_figure : mpl.figure.Figure | None = None, 
			pf_gridspec : mpl.gridspec.GridSpec | None = None, 
			gridspec_index : int = 0, 
			window_title :str = 'Image Viewer'
		):
		
		
		self.parent_figure = plt.figure(figsize=(15,10), layout='constrained') if parent_figure is None else parent_figure
		self.pf_gridspec = self.parent_figure.add_gridspec(1,1,left=0,right=0,top=1,bottom=1,wspace=0,hspace=0) if pf_gridspec is None else pf_gridspec
		self.figure = self.parent_figure.add_subfigure(self.pf_gridspec[gridspec_index], frameon=False)
		self.set_window_title(window_title)
		
		self.f_gridspec = self.figure.add_gridspec(2,2,left=0.4,right=0.6,top=0.7,bottom=0.3,wspace=0,hspace=0)
		
		
		self.input_image_viewer = ImageViewer(self.figure, pf_gridspec=self.f_gridspec, gridspec_index=0, window_title=None)
		self.ssa_evalue_panel = LineRegionPanel(self.figure, self.f_gridspec, 1, None)
		self.ssa_aggregate_viewer = ImageAggregator(self.figure, pf_gridspec=self.f_gridspec, gridspec_index=2, window_title=None)
		self.ssa_residual_viewer = ResidualViewer(self.figure, pf_gridspec=self.f_gridspec, gridspec_index=3, window_title=None)
		
		
		self.input_image_viewer.main_axes_im.set_title('Input Data')
		
		
		self.ssa_evalue_panel.main_axes.set_yscale('log')
		self.ssa_evalue_panel.main_axes.set_xlabel('Singular Value Index')
		self.ssa_evalue_panel.main_axes.set_ylabel('Singular Value')
		self.ssa_evalue_panel.main_plot_handle.set_title('SSA Singular Values of Components')
		self.ssa_evalue_panel.set_region_limits(0,None,1,1,None)
		self.ssa_evalue_panel.attach_callback('on_region_changed',
			self.set_image_plane_range_from_evalue_panel_region
		)
		
		self.ssa_aggregate_viewer.main_axes_im.set_title('Sum of SSA Components')
		self.ssa_aggregate_viewer.attach_callback('on_set_image_agg_planes',
			self.set_image_plane_range_from_aggregator
		)
		
		self.ssa_residual_viewer.main_axes_im.set_title('Residual (Input Data - Sum of SSA Components)')
		
		
		self.save_button = Button(self.figure, (0.05, 0.001, 0.15, 0.03), 'Save Aggregated SSA Components')
		self.save_button.on_clicked(self.save_aggregated_ssa_components)		
		self.save_path_text = TextBox(self.figure, (0.05, 0.031, 0.3, 0.02), label='path')
		self.save_checks = CheckButtons(self.figure, (0.2,0.001, 0.15,0.03), ['overwrite', 'create parent folders'])
		
		
		
		self.save_feedback_text = self.figure.text(0.05, 0.052, 'Formats: ".npy", ".tiff"')
		
		
		
		
		self.input_data = None
		self.ssa_data = None


	def save_aggregated_ssa_components(self, event):
		save_path = Path(self.save_path_text.get_value())
		_lgr.debug(f'{save_path=}')
		can_overwrite, can_create_parents = self.save_checks.get_value()
		
		success_flag = False
		feedback = [f'Saving to "{save_path.absolute()}"']
		fail_feedback = []
		success_feedback = []
		
		_lgr.debug(f'Collecting failure feedback')
		if save_path.exists():
			fail_feedback.append('DESTINATION ALREADY EXISTS')
			if save_path.is_dir():
				fail_feedback.append('AND IS A DIRECTORY') 
			elif save_path.is_file():
				fail_feedback.append('AND IS A FILE')
		elif not save_path.parent.exists():
			fail_feedback.append('PARENT DIRECTORY NOT PRESENT')
		
		ext = save_path.suffix
		
		
		if (not save_path.exists() or (can_overwrite and not save_path.is_dir())) and (save_path.parent.exists() or can_create_parents):
			_lgr.debug(f'Starting save attempt...')
			try:
				if not save_path.parent.exists() and can_create_parents:
					save_path.parent.mkdir(parents=True, exist_ok=True)
				
				data = self.ssa_aggregate_viewer.get_displayed_data()
				match ext:
					case '.npy':
						np.save(save_path, data)
					case '.tiff':
						PIL.Image.fromarray(data).save(save_path)
					case _:
						raise RuntimeError(f'Unknown extension "{ext}"')
			except Exception as e:
				fail_feedback.append(str(e))
				success_flag=False
			else:
				success_feedback.append('successful')
				success_flag = True
			_lgr.debug(f'Completed save attempt.')
		
		_lgr.debug(f'Writing feedback...')
		if success_flag:
			self.save_feedback_text.set_text(' '.join(feedback+success_feedback))
			self.save_feedback_text.set_color('green')
		else:
			self.save_feedback_text.set_text(' '.join(feedback+fail_feedback))
			self.save_feedback_text.set_color('red')
		_lgr.debug(f'Feedback written.')
		return True

	def set_image_plane_range_from_evalue_panel_region(self, r_min : int, r_max : int):
		if r_min is None:
			r_min = 0
		if r_max is None:
			r_max = len(self.ssa_evalue_panel.main_data_handle)
		self.ssa_aggregate_viewer.image_plane_slider.set_value((r_min,r_max))
		self.ssa_aggregate_viewer.main_axes_im.set_data(np.nansum(self.ssa_aggregate_viewer.main_axes_image_data[int(r_min):int(r_max)], axis=0))
		self.set_residual_data(self.input_image_viewer.get_displayed_data(),self.ssa_aggregate_viewer.get_displayed_data())

	def set_image_plane_range_from_aggregator(self, agg : tuple[int,int]):
		x = agg.image_plane_slider.get_value()
		self.set_residual_data(self.input_image_viewer.get_displayed_data(),agg.get_displayed_data())
		self.ssa_evalue_panel.set_region(x[0],x[1]+1)

	def set_residual_data(self, obs_data : np.ndarray, model_data : np.ndarray):
		residual = obs_data - model_data
		self.ssa_residual_viewer.frozen_clims = (
			min(np.nanmin(obs_data), np.nanmin(residual)),
			max(np.nanmax(obs_data), np.nanmin(residual))
		)
		self.ssa_residual_viewer.set_data(residual)
		

	def set_window_title(self, title : str):
		if title is not None:
			self.figure.canvas.manager.set_window_title(title)

	def show(self, data : np.ndarray = None, title : str = None):
		"""
		Show the SSA decomposition of `data` with `title`
		"""
		plt.figure(self.parent_figure)
		
		self.set_data(data)
		
		plt.show()
	
	def set_data(self, data : np.ndarray):
		self.input_data = data
		
		self.input_image_viewer.set_data(self.input_data)
		
		self.ssa = SSA(
			self.input_image_viewer.get_displayed_data(),
			(5,5),
			rev_mapping='fft',
			svd_strategy='eigval', # uses less memory and is faster
			#grouping={'mode':'elementary'}
			grouping={'mode':'similar_eigenvalues', 'tolerance':0.01}
		)
		_lgr.debug(f'{self.ssa.X_ssa.shape=} {self.ssa.m=} {np.diag(self.ssa.s_g)=}')
		self.ssa_evalue_panel.set_data(np.diag(self.ssa.s_g))
		self.ssa_aggregate_viewer.set_data(self.ssa.X_ssa)
		self.ssa_residual_viewer.set_data(self.input_image_viewer.get_displayed_data() - self.ssa_aggregate_viewer.get_displayed_data())

if __name__ == '__main__':
	#image = sys.argv[1]
	
	if len(sys.argv) < 2:
		image = np.sqrt(np.sum(((np.indices((20, 100, 128)).T - np.array([10,50,64])).T)**2, axis=0))
		for i in range(image.shape[0]):
			image[i,...] = image[i]**(2*((i+1)/image.shape[0]))
	else:
		image = np.array(PIL.Image.open(sys.argv[1]))
		
		# Can only use greyscale images, so sum along colour axis
		if image.ndim == 3:
			image = np.sum(image, axis=-1)
	
	
	
	
	imviewer = SSAViewer()
	imviewer.show(image, 'image')


"""
Defines the public facing interface for plot classes, used to interface with hooks in aopp_deconv_tool.algorithm.deconv classes
"""
import dataclasses as dc
from collections import namedtuple
from typing import Any, Callable, ClassVar
from functools import wraps

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt


import aopp_deconv_tool.plot_helper as plot_helper
#AxisDataMapping = namedtuple('AxisDataMapping', ('label','attribute','limit_getter'))

@dc.dataclass(repr=False, eq=False, slots=True)
class AxisDataMapping:
	"""
	Defines the relationship between a "datasource" attribute and the axis that it is plotted along
	"""
	label : str
	attribute : str | None = None
	limit_getter : Callable[[Any],tuple[float,float]] = plot_helper.lim
	
	def get_limits(self, d):
		return self.limit_getter(d)

@dc.dataclass(repr=False, eq=False, slots=True)
class Base:
	
	# independent and dependent variables in the format
	# {'label1':'attribute1', 'label2':'attribute2',...}
	# a specific plot type has predefined labels and attributes, these can be
	# overwritten in the __init__ of the class
	title : str = 'Base Plot Helper Title'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_labels : tuple[str,...] = tuple()
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :
		tuple() # (('ax1_label','ax1_attr'), ('ax2_label','ax2_attr'), ...)
	)
	
	static_frame : bool = True
	plt_kwargs : dict = dc.field(default_factory=dict)
	ax_funcs : list = dc.field(default_factory=list)
	show_limits_in_title : dict = dc.field(default_factory=dict) #{0:False, 1:False, 2:False}, 0,1,2 are axis numbers (x,y,z)
	
	
	# Internal attributes
	# Will be assigned after initialisation
	datasource_name : str = dc.field(default='datasource name', init=False)
	ax : mpl.axes.Axes = dc.field(default=None, init=False) # matplotlib axis
	datasource : Any = dc.field(default = None, init=False)
	datasource_data_getter : Callable | None = dc.field(default=None, init=False)
	
	hdl : Any = dc.field(default=None,init=False) # handle to visualisation of the plot
	n_updates : int = dc.field(default=0, init=False)

	
	def get_axis_limits(self,ax_num):
		match ax_num:
			case 0:
				return self.ax.get_xlim() if self.ax is not None else None
			case 1:
				return self.ax.get_ylim() if self.ax is not None else None
			case 2:
				return self.hdl.get_clim() if self.hdl is not None else None
			case _:
				raise RuntimeError(f'Plot {self.title=} does not have an axis of number {ax_num}')
	
	def get_axis_limits_str(self, ax_num):
		al = self.get_axis_limits(ax_num)
		if al is None:
			return 'None'
		return "{0:5.3g},{1:5.3g}".format(*al)
	
	def set_ax_title(self):
		if self.title is not None:
			axis_limit_strs = []
			for ax_num, show_limits_flag in self.show_limits_in_title.items():
				if show_limits_flag:
					axis_limit_strs.append(f'{self.axis_data_mappings[ax_num].label}:[{self.get_axis_limits_str(ax_num)}]')
			return self.ax.set_title(self.title
				+ (('\n' + ' '.join(axis_limit_strs)) if len(axis_limit_strs)>0 else '')
			)
		return None
	
	def is_datasource_attached(self):
		return hasattr(self, 'datasource') and self.datasource is not None
	
	def on_attach_datasource(self):
		"""
		Performs any initial setup that is required when a datasource is attached.
		Should be overwritten by subclass.
		"""
		self.n_updates=0
		if not self.is_datasource_attached():
			raise RuntimeError(f'{self} does not have a "datasource" attribute')
	
	def attach_datasource(self, datasource, datasource_data_getter):
		self.datasource = datasource
		self.datasource_data_getter = datasource_data_getter
		self.on_attach_datasource()
		return(self)
	
	def is_ax_attached(self):
		return hasattr(self, 'ax') and self.ax is not None
	
	def on_attach_ax(self):
		"""
		Performs any initial setup that is required when an axes is attached.
		Should be overwritten by subclass
		"""
		if not self.is_ax_attached():
			raise RuntimeError(f'{self} does not have a "axes" attribute')
	
	def attach_ax(self, ax):
		self.ax = ax
		self.set_ax_title()
		
		for i, (set_axis_label, adm) in enumerate(zip((
					self.ax.set_xlabel, 
					self.ax.set_ylabel
				), 
				self.axis_data_mappings
			)):
			if adm.label is not None:
				if len(self.axis_labels) > i and self.axis_labels[i] is not None:
					set_axis_label(self.axis_labels[i])
				else:
					set_axis_label(adm.label)
		for ax_func in self.ax_funcs:
			ax_func(self.ax)
		self.on_attach_ax()
		return(self)
	
	
	def attach(self, ax : mpl.axes.Axes, datasource : Any, datasource_data_getter : Callable[[Any],tuple[Any]] ):
		"""
		Attach a `datasource` (input) and an axes `ax` (output) to the plot.
		"""
		self.attach_datasource(datasource, datasource_data_getter)
		self.attach_ax(ax)
		return(self)

	def detach_datasource(self):
		self.datasource = None
	
	def detach_ax(self):
		self.ax = None
		
	def detach(self):
		self.detatch_datasource()
		self.detatch_ax()

	def update_plot_data(self, data):
		raise NotImplementedError
	
	
	def update_plot_visual(self):
		for set_lims, adm in zip((
					self.ax.set_xlim,
					self.ax.set_ylim,
					self.hdl.set_clim if hasattr(self.hdl,'set_clim') else lambda *a, **k: None
				), 
				self.axis_data_mappings
			):
			if adm.attribute is not None:
				d = getattr(self, adm.attribute)
				set_lims(*adm.get_limits(d))
		
		if not self.static_frame:
			self.set_ax_title()
			if not self.ax.drawn:
				for x in self.ax.get_children():
					self.ax.draw_artist(x)
				self.ax.drawn = True
					
		self.ax.draw_artist(self.hdl)
		
	def update(self):
		assert self.is_datasource_attached(), "requires datasource attribute is not None"
		
		self.update_plot_data(self.datasource_data_getter(self.datasource))
		self.update_plot_visual()
		self.n_updates += 1
	
	def iter_ax(self, attrs):
		assert self.is_ax_attached(), "require ax attribute"
		for attr in attrs:
			yield getattr(self.ax, attr)
	
	
	


			



"""
Tool for interpolating data in a FITS file, pixels to be interpolated are specified via a boolean map.
"""


import sys
from pathlib import Path
import dataclasses as dc
from typing import Literal, Callable, NewType
import functools

import numpy as np
import scipy as sp
from astropy.io import fits

from aopp_deconv_tool.task_strat import TaskStratInfo, TaskStratSet

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
from aopp_deconv_tool.astropy_helper.fits.specifier import FitsSpecifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.numpy_helper.array.grid
import aopp_deconv_tool.arguments

import aopp_deconv_tool.scipy_helper as sph
import aopp_deconv_tool.scipy_helper.interp
import aopp_deconv_tool.scipy_helper.label_ops

from aopp_deconv_tool.algorithm.interpolate.ssa_interp import ssa_interpolate_at_mask, ssa_deviations_interpolate_at_mask
from aopp_deconv_tool.algorithm.bad_pixels.ssa_sum_prob import ssa2d_sum_prob_map

from aopp_deconv_tool.py_ssa import SSA

import matplotlib.pyplot as plt

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


type NM = NewType('NM', int)


def array_ssa_deviations_interpolate_at_mask(a : np.ndarray, mask : np.ndarray, **kwargs):
	return ssa_deviations_interpolate_at_mask(SSA(a, **kwargs), mask)

def array_ssa_interpolate_at_mask(a : np.ndarray, mask : np.ndarray, **kwargs):
	return ssa_interpolate_at_mask(a, SSA(a, **kwargs), mask)


@dc.dataclass
class InterpolationStratInfo (TaskStratInfo):
	"""
	Holds information (name, description) about a callable that is used to interpolate a numpy array at it's first argument, at pixels specified by a numpy array as it's second argument, and return the resulting numpy array.
	"""
	callable : Callable[[np.ndarray['NM',float], np.ndarray['NM',bool]], np.ndarray['NM',float]]


# A list of strategies that can be used to perform the "interpolation" task.
interpolation_strategies = TaskStratSet(
'Performs interpolation upon an array at positions provided by a mask'
).add(
	InterpolationStratInfo(
		'scipy',
		'Uses scipy routies to interpolate over the masked regions, edge effects are minimised by convolving with a 3x3 square before interpolation',
		functools.partial(sph.interp.interpolate_at_mask, edges='convolution')
	),
	InterpolationStratInfo(
		'ssa',
		'[EXPERIMENTAL] Calculates singular spectrum analysis (SSA) components, and uses the first 25% of them to fill in masked regions.',
		functools.partial(array_ssa_interpolate_at_mask, w_shape=10, grouping = {'mode':'elementary'}),
	),
	InterpolationStratInfo(
		'ssa_deviation',
		'[EXPERIMENTAL] Within a masked region, replaces pixels of singular spectrum analysis (SSA) components with high deviations with median values, which are summed and replaced into the masked region on the original image.',
		functools.partial(array_ssa_deviations_interpolate_at_mask, w_shape=10, grouping = {'mode':'elementary'}),
	)
)




def run(
		data_fits_spec : FitsSpecifier,
		bpmask_fits_spec : FitsSpecifier,
		output_path : Path,
		interp_method : str = 'scipy',
	):
	
	original_data_type=None
	
	with fits.open(Path(data_fits_spec.path)) as data_hdul, fits.open(Path(bpmask_fits_spec.path)) as bpmask_hdul:
		
		
		_lgr.debug(f'{data_fits_spec}')
		_lgr.debug(f'{bpmask_fits_spec}')
		
		
		
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[data_fits_spec.ext]
		data = data_hdu.data
		original_data_type=data_hdu.data.dtype
		
		bpmask_hdu = bpmask_hdul[bpmask_fits_spec.ext]
		if issubclass(bpmask_hdu.data.dtype.type, np.integer):
			bpmask = bpmask_hdu.data.astype(bool)
		elif issubclass(bpmask_hdu.data.dtype.type, np.bool_):
			bpmask = bpmask_hdu.data
		else:
			raise RuntimeError(f'Bad Pixel Mask "{bpmask_fits_spec.path}" is not a boolean mask. It is of type "{bpmask_hdu.data.dtype}", only integer types can be converted to a boolean mask.')
		
		# Ensure that the data and mask are compatible shapes
		data_shape = data[data_fits_spec.slices].shape
		bpmask_shape = bpmask[bpmask_fits_spec.slices].shape
		if (len(data_shape) != len(bpmask_shape)) or any(s1!=s2 for s1,s2 in zip(data_shape, bpmask_shape)):
			raise RuntimeError(f'data and bad pixel mask are of incompatible shape {data_shape} vs {bpmask_shape} with {data_fits_spec=} {bpmask_fits_spec=}')
			
		# Allocate array to hold interpolated data
		interp_data = np.full_like(data, fill_value=np.nan)
		residual = np.full_like(data, fill_value=np.nan)
		
		# Loop over the index range specified by `obs_fits_spec` and `psf_fits_spec`
		for i, idx in enumerate(nph.slice.iter_indices(data, data_fits_spec.slices, data_fits_spec.axes['CELESTIAL'])):
			_lgr.debug(f'{i=} current_idx={idx[0][tuple(0 for i in data_fits_spec.axes["CELESTIAL"])]}')
			
			interp_data[idx] = interpolation_strategies(interp_method, data[idx], bpmask[idx] | np.isnan(data[idx]) | np.isinf(data[idx]))
			residual[idx] = data[idx] - interp_data[idx]
			
	
	
		hdr = data_hdu.header
		param_dict = {
			'original_file' : Path(data_fits_spec.path).name, # record the file we used
			'bpmask_file' : Path(bpmask_fits_spec.path).name, 
			'interp_method' : interp_method,
		}

		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='interpolate',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
				

	
	# Save the products to a FITS file
	hdu_interp = fits.PrimaryHDU(
		header = hdr,
		data = interp_data.astype(original_data_type)
	)
	hdu_residual = fits.ImageHDU(
		header = hdr,
		data = residual.astype(original_data_type),
		name='RESIDUAL'
	)
	hdul_output = fits.HDUList([
		hdu_interp,
		hdu_residual
	])
	hdul_output.writeto(output_path, overwrite=True)
	

def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_interp'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter,
		epilog=FITS_SPECIFIER_HELP
	)
	
	parser.add_argument(
		'data_fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon . See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	parser.add_argument(
		'bpmask_fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon. See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	#parser.add_argument('-o', '--output_path', help=f'Output fits file path. By default is same as the `data_fits_spec` path with "{DEFAULT_OUTPUT_TAG}" appended to the filename')
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `data_fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	
	parser.add_argument('--interp_method', choices=interpolation_strategies.names, default=interpolation_strategies.names[0], help=interpolation_strategies.format_description())


	args = parser.parse_args(argv)
	
	args.data_fits_spec = aph.fits.specifier.parse(args.data_fits_spec, DESIRED_FITS_AXES)
	args.bpmask_fits_spec = aph.fits.specifier.parse(args.bpmask_fits_spec, DESIRED_FITS_AXES)
	
	#if args.output_path is None:
	#	fpath = Path(args.data_fits_spec.path)
	#	args.output_path =  (fpath.parent / (str(fpath.stem)+DEFAULT_OUTPUT_TAG+str(fpath.suffix)))
	other_file_path = Path(args.data_fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	return args

def go(
		data_fits_spec,
		bpmask_fits_spec, 
		output_path=None,
		interp_method=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# Add stuff to kwargs here if needed
	
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=2)
	exec_with_args(arglist)

def exec_with_args(argv):
	args = parse_args(argv)
	
	run(
		args.data_fits_spec, 
		args.bpmask_fits_spec,
		output_path=args.output_path,
		interp_method=args.interp_method, 
	)


if __name__ == '__main__':
	exec_with_args(sys.argv[1:])
	



"""
Copy the WCS information from one FITS file to another, useful when we know that two observations were taken with the same setup but one has no physical information.
"""



import sys, os
from pathlib import Path


from astropy.io import fits
from astropy.wcs import WCS

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.astropy_helper.fits.specifier
from aopp_deconv_tool.astropy_helper.fits.specifier import FitsSpecifier

from aopp_deconv_tool.fpath import FPath
from aopp_deconv_tool.numpy_helper.axes import AxesOrdering
import aopp_deconv_tool.arguments

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')



def get_fits_file_type(hdr):
	"""
	Return the type of a fits file, or `None` if type is not recognised.
	"""
	if hdr.get('HDUCLASS', None) == 'ESO':
		return 'ESO'
	return None


def run(
		donor_fits_spec : FitsSpecifier,
		acceptor_fits_spec : FitsSpecifier,
		output_path : Path | str
	):
	_lgr.debug('DEBUGGING')
	
	
	with fits.open(Path(donor_fits_spec.path)) as donor_hdul, fits.open(Path(acceptor_fits_spec.path)) as acceptor_hdul:
		donor_hdr = donor_hdul[donor_fits_spec.ext].header
		acceptor_hdr = acceptor_hdul[acceptor_fits_spec.ext].header
		
		# Want to make the resulting cube as close to "true" as possible, therefore use different strategies for different datasets
		donor_fits_file_type = get_fits_file_type(donor_hdr)
		acceptor_fits_file_type = get_fits_file_type(acceptor_hdr)
		
		donor_header_keys = aph.fits.header.get_wcs_header_keys(donor_hdr, donor_fits_spec.axes['ALL'])
		acceptor_header_keys = aph.fits.header.get_wcs_header_keys(acceptor_hdr, donor_fits_spec.axes['ALL'])
		
		_lgr.debug('donor_header_keys')
		for k,v in donor_header_keys.items():
			_lgr.debug(f'{k} = {v}')
		
		_lgr.debug('acceptor_header_keys')
		for k,v in acceptor_header_keys.items():
			_lgr.debug(f'{k} = {v}')
		
		# Copy donor header keys to acceptor header keys
		for k,v in donor_header_keys.items():
			if v is not None:
				acceptor_header_keys[k] = v
		
		
		match acceptor_fits_file_type:
			case 'ESO':
				# ESO files have the primary extension be information about the dataset, and the others hold the image data
				eso_dataset_hdr = acceptor_hdul['PRIMARY'].header
				eso_dataset_ra = eso_dataset_hdr.get('RA',None)
				eso_dataset_dec = eso_dataset_hdr.get('DEC',None)
				
				# Replace the starting points of the RA,DEC axes with these values
				for k in acceptor_header_keys:
					if 'CTYPE' in k:
						if 'RA' in acceptor_header_keys[k]:
							CRVAL_key = 'CRVAL'+k[5:]
							acceptor_header_keys[CRVAL_key] = eso_dataset_ra
						elif 'DEC' in acceptor_header_keys[k]:
							CRVAL_key = 'CRVAL'+k[5:]
							acceptor_header_keys[CRVAL_key] = eso_dataset_dec
						
		# update acceptor header with new keys, only update keys that are present
		for k,v in acceptor_header_keys.items():
			if (v is not None) and (k in acceptor_hdr):
				_lgr.debug(f'{k} = {v}')
				acceptor_hdr[k] = v
		
		
		acceptor_hdr.update(aph.fits.header.DictReader(
			{
				'original_file' : acceptor_fits_spec.path,
				'donor_file' : donor_fits_spec.path,
			},
			prefix='copy_wcs',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(acceptor_hdr)
		))
		
		hdu_wcs_copy = fits.PrimaryHDU(
			header = acceptor_hdr,
			data = acceptor_hdul[acceptor_fits_spec.ext].data
		)
		hdul_output = fits.HDUList([
			hdu_wcs_copy,
		])
		hdul_output.writeto(output_path, overwrite=True)
				
			
	





def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_wcs_paired'
	DESIRED_FITS_AXES = ['ALL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	class ArgFormatter (argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter):
		def __init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=ArgFormatter,
		epilog=FITS_SPECIFIER_HELP,
		exit_on_error=False
	)
	
	parser.add_argument(
		'donor_fits_spec',
		help = 'FITS file specifier, will use the WCS data from this file',
		type=str,
		metavar='FITS Specifier',
	)
	
	parser.add_argument(
		'acceptor_fits_spec',
		help = 'FITS file specifier, will change the WCS data in this file',
		type=str,
		metavar='FITS Specifier',
	)
	
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `obs_fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)

	
	args = parser.parse_args(argv)
	
	args.donor_fits_spec = aph.fits.specifier.parse(args.donor_fits_spec, DESIRED_FITS_AXES)
	args.acceptor_fits_spec = aph.fits.specifier.parse(args.acceptor_fits_spec, DESIRED_FITS_AXES)
	
	other_file_path = Path(args.acceptor_fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	return args


def go(
		donor_fits_spec,
		acceptor_fits_spec,
		output_path=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# Add stuff to kwargs here if needed
	
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=2)
	exec_with_args(arglist)

def exec_with_args(argv):
	args = parse_args(argv)
	_lgr.info('#### ARGUMENTS ####')
	for k,v in vars(args).items():
		_lgr.info(f'\t{k} : {v}')
	_lgr.info('#### END ARGUMENTS ####')
	
	run(
		args.donor_fits_spec, 
		args.acceptor_fits_spec,
		args.output_path, 
	)

if __name__ == '__main__':
	exec_with_args(sys.argv[1:])


"""
Dependency injectors for each psf model, consistent interface means they work well with psf_data_ops
"""

from aopp_deconv_tool.psf_model.dependency_injector.radial_psf_model_di import RadialPSFModelDependencyInjector
from aopp_deconv_tool.psf_model.dependency_injector.adaptive_optics_psf_model_di import MUSEAdaptiveOpticsPSFModelDependencyInjector
from aopp_deconv_tool.psf_model.dependency_injector.gaussian_psf_model_di import GaussianPSFModelDependencyInjector
from aopp_deconv_tool.psf_model.dependency_injector.turbulence_psf_model_di import TurbulencePSFModelDependencyInjector




from aopp_deconv_tool.psf_model.dependency_injector.params_and_psf_model_di import *#ParamsAndPsfModelDependencyInjector
from aopp_deconv_tool.psf_model.radial_psf_model import RadialPSFModel


class RadialPSFModelDependencyInjector(ParamsAndPsfModelDependencyInjector):
	"""
	Models the PSF as a radial histogram with `nbins` from point (`x`,`y`)
	"""
	
	def __init__(self, psf_data):
		
		super().__init__(psf_data)
		
		self._params = PriorParamSet(
			PriorParam(
				'x',
				(0, psf_data.shape[0]),
				False,
				psf_data.shape[0]//2,
				"centre point of radial histogram on x-axis"
			),
			PriorParam(
				'y',
				(0, psf_data.shape[1]),
				False,
				psf_data.shape[1]//2,
				"centre point of radial histogram on y-axis"
			),
			PriorParam(
				'nbins',
				(0, np.inf),
				True,
				50,
				"Number of bins in radial histogram"
			)
		)
		
		self._psf_model = RadialPSFModel(
			psf_data
		)
		
	
	def get_parameters(self):
		return self._params
	
	def get_psf_model_flattened_callable(self): 
		return self._psf_model
	
	def get_psf_result_postprocessor(self): 
		def psf_result_postprocessor(params, psf_model_flattened_callable, fitted_vars, consts):
			params.apply_to_callable(
				psf_model_flattened_callable, 
				fitted_vars,
				consts
			)
			return psf_model_flattened_callable.centreed_result
			
		return psf_result_postprocessor


"""
Field-aware path class, can use format fields to create new paths based upon a template path.
"""

from pathlib import PurePath, Path
import re

class FPath(PurePath):
	"""
	A field-aware abstract path (PurePath). Instantiate via FPath('path/{with}_some_{field}'). String used in instantiation will be passed to str.format(...) when used.
	
	Defaults are set via keyword arguments on construction.
	
	Use the `FPath.with_fields(**kwargs)` method to create a concrete path (Path) with **kwargs passed to str.format(...). These **kwargs overwrite defaults.
	
	"""
	
	field_re = re.compile(r'{{1}(.*?)(:*?)?}{1}', flags=re.S) # Regular expression that finds field names in passed path strings
	
	def __init__(self, *args, **kwargs):
		super().__init__(*args)
		self._fields = []
		self._defaults = {}
		self._parts = []
		for a in args:
			self._parts.extend(PurePath(a).parts if (type(a) in (str, bytes)) else a.parts)
		self._parts = tuple(self._parts)
		
		for p in self.parts:
			if type(p) is str:
				for match in FPath.field_re.finditer(p):
					self._fields.append(match.group(1))
		self._fields = tuple(set(self._fields))
		
		# update only the fields present so we don't carry around unused data
		for field in self.fields:
			if field in kwargs:
				self._defaults[field] = kwargs[field]
	
	@property
	def fields(self):
		"""
		Return the template fields of the FPath
		"""
		return self._fields
	
	@property
	def defaults(self):
		"""
		Return the default values of the FPath's template fields
		"""
		return self._defaults
	
	@defaults.setter
	def defaults(self, **kwargs):
		"""
		Set the default values of the FPath's template fields
		"""
		self._defaults = dict((field, kwargs[field]) for field in self.fields)
	
	def with_fields(self, **kwargs):
		"""
		Return a new Path object created by substituting the values in `**kwargs` for the FPath's template fields. Any template fields not in `**kwargs` will take their default values.
		"""
		temp_dict = {}
		temp_dict.update(self._defaults)
		temp_dict.update(kwargs)
		try:
			return Path(*(p.format(**temp_dict) for p in self._parts))
		except KeyError as e:
			e.add_note(f'All fields of FPath must have a default or set a value using `FPath.with_fields(**kwargs)`. Fields with no value for this instance: {', '.join((f'"{f}"' for f in self.fields if f not in temp_dict))}')
			raise e
	
	def __fspath__(self):
		"""
		Return the "filesystem-path" of the current FPath object, all fields will take their default values.
		"""
		return self.with_fields().__fspath__()


"""
Convert an image to FITS format
"""

import sys, os
from pathlib import Path
import dataclasses as dc
import argparse
import io
import re
from typing import Any

from collections import namedtuple

from astropy.io import fits
import PIL
import PIL.Image
import PIL.features
import PIL.ExifTags

import matplotlib.pyplot as plt
import numpy as np

from aopp_deconv_tool import plot_helper
import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.header
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.arguments

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')

# DataBundle is a grouping of related data and header information
DataBundle = namedtuple('DataBundle', ('data', 'header'))

re_dashed_line = re.compile(r'\n-+\n') # lines just consisting of "-" characters
re_comma_space = re.compile(r',\s') # comma then a space

def get_supported_formats():
	"""
	Return a list of supported file extensions that can be successfully converted to FITS format
	"""
	ss = io.StringIO()
	PIL.features.pilinfo(out = ss, supported_formats=True)
	
	supported_extensions = []
	for chunk in re_dashed_line.split(ss.getvalue()):
		_lgr.debug(f'{chunk=}')
		extensions_are_supported = False
		possibly_supported_extensions = []
		
		for line in chunk.split('\n'):
			if line.startswith('Extensions: '):
				possibly_supported_extensions = re_comma_space.split(line[12:])
				_lgr.debug(f'{possibly_supported_extensions=}')
			if line.startswith('Features: '):
				if 'open' in line[10:]:
					extensions_are_supported = True
		if extensions_are_supported:
			supported_extensions.extend(possibly_supported_extensions)
			
	return supported_extensions


def read_exif(image, header={}, exif_tag_reader=lambda k, v: (str(k), v)):
	"""
	Read the EXIF tags of an image
	"""
	exif = image.getexif()
	_lgr.debug(f'{exif=}')
	for k,v in exif.items():
		tag, value = exif_tag_reader(k,v)
		header[tag] = value
		_lgr.debug(f'{k} : {v}')
	_lgr.debug(f'DONE')
	return header

def read_image_into_data_bundle(fpath : Path) -> DataBundle:
	"""
	Read an image at `fpath` into a numpy array. Store EXIF tags in the header field of a DataBundle, and the image data in "data" field of a DataBundle.
	"""
	match fpath.suffix:
		case '.tif' | '.tiff':
			exif_tag_reader = lambda k, v: (PIL.TiffTags.lookup(k).name, v)
		case _:
			exif_tag_reader=lambda k, v: (str(k), v)
	
	
	data = None	
	header = {}
	with PIL.Image.open(fpath) as image:
		header = read_exif(image, header=header, exif_tag_reader=exif_tag_reader)
		image = image.convert(mode='F')
		data = np.array(image).astype(np.float64)
		
	return DataBundle(data, header)

def plot_image_and_fits(image_path : str | Path, fits_path : str | Path, ext=None, slices=None, fig=None, ax=None, fig_kwargs : dict[str,Any] = {'figsize':(12,4)}):
	"""
	Uses matplotlib to display an image and a FITS file side-by-side
	"""
	
	image = read_image_into_data_bundle(image_path)
	fits_header = fits.getheader(fits_path, ext=ext)
	fits_data = fits.getdata(fits_path, ext=ext)
	
	# Note, can use slices=(0,slice(None),slice(None)) to reduce number of dimensions
	if slices is not None:
		fits_data = fits_data[*slices]
	
	assert fits_data.ndim == 2, f"{fits_path} does not have exactly 2 dimensions when read in. Need to slice or index the file."
	
	
	fig, ax = plot_helper.ensure_fig_and_ax(fig, ax, fig_kwargs, (1,3))
	fig.suptitle('Image and FITS data with data range.')
	
	ax[0].set_title(f'Image data [{np.nanmin(image.data):+9.2E},{np.nanmax(image.data):+9.2E}]')
	ax[0].imshow(image.data)
	ax[0].set_axis_off()
	
	ax[1].set_title(f'FITS data [{np.nanmin(fits_data):+9.2E},{np.nanmax(fits_data):+9.2E}]')
	ax[1].imshow(fits_data)
	ax[1].set_axis_off()
	
	residual = image.data - fits_data
	ax[2].set_title(f'Image - FITS [{np.nanmin(residual):+9.2E},{np.nanmax(residual):+9.2E}]')
	ax[2].imshow(image.data - fits_data)
	ax[2].set_axis_off()
	
	return fig, ax
	
	
	
	
	
	


def save_as_fits(output_path : str | Path, primary_data_bundle : DataBundle, **kwargs : dict[str, DataBundle]):
	"""
	Save a `primary_data_bundle` and a list of other data bundles to a specified path
	"""
	hdr = fits.Header()
	update_if_not_present = lambda hdr, key, value: hdr.__setitem__(key, value if key not in hdr else hdr[key])
	# If we do not already have CTYPEs for axes, we need to add them here
	# assume axis 0 is spectral, axis 1 is celestial, axis 2 is celestial
	ctypes = ('SPEC_PIX', 'SKY__PIX', 'SKY__PIX')
	cunits = ('PIXEL', 'PIXEL', 'PIXEL')
	for i in range(len(ctypes)):
		ctype_i_key = f'CTYPE{3 - i}' # convert from numpy order to FITS order
		cunit_i_key = f'CUNIT{3 - i}' # convert from numpy order to FITS order
		crval_i_key = f'CRVAL{3 - i}' # convert from numpy order to FITS order
		cdelt_i_key = f'CDELT{3 - i}' # convert from numpy order to FITS order
		crpix_i_key = f'CRPIX{3 - i}' # convert from numpy order to FITS order
		update_if_not_present(hdr, ctype_i_key, ctypes[i])
		update_if_not_present(hdr, cunit_i_key, cunits[i])
		update_if_not_present(hdr, crval_i_key, 1)
		update_if_not_present(hdr, cdelt_i_key, 1)
		update_if_not_present(hdr, crpix_i_key, 1)
			
	hdr.update(aph.fits.header.DictReader(primary_data_bundle.header))
	
	# Other routines rely on 3 axes, so make our output have 3 axes
	if primary_data_bundle.data.ndim == 2:
		output_data = primary_data_bundle.data[None,...]
	else:
		output_data = primary_data_bundle.data
	
	
	
	
	hdu_primary = fits.PrimaryHDU(
		header = hdr,
		data = output_data
	)
	
	hdu_list = [hdu_primary]
	
	for k,v in kwargs.items():
		new_hdr = fits.Header()
		new_hdr.update(aph.fits.header.DictReader(v.header))
		hdu_list.append(fits.ImageHDU(header=new_hdr, name=k, data=v.data))
	
	
	hdul = fits.HDUList(hdu_list)
	
	hdul.writeto(output_path, overwrite=True)
	_lgr.info(f'Converted file written to "{output_path}"')


def parse_args(argv):
	parser = argparse.ArgumentParser(
		description=__doc__,
	)
	
	SUPPORTED_FORMATS = get_supported_formats()
	_lgr.debug(f'{SUPPORTED_FORMATS=}')
	
	parser.add_argument('image_path', type=Path, help=f'Path to the image to convert, can be one of {" ".join(SUPPORTED_FORMATS)}')
	#parser.add_argument('-o', '--output_path', type=Path, default=None, help='Path to save fits conversion to, if not supplied will replace original file extension with ".fits"')
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}.fits',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `image_path` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	
	args = parser.parse_args(argv)
	
	if args.image_path.suffix not in SUPPORTED_FORMATS:
		parser.print_help()
		_lgr.error(f'Unsupported image format {args.image_path.suffix}')
		sys.exit()
	
	#if args.output_path is None:
	#	args.output_path = Path(str(args.image_path.stem) + '.fits')
	other_file_path = Path(args.image_path)
	args.output_path = args.output_path.with_fields(
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	for k,v in vars(args).items():
		_lgr.debug(f'{k} = {v}')
	
	return args



def go(
		image_path,
		output_path=None,
		help_FLAG=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""	
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args = parse_args(argv)
	
	
	data_bundle = read_image_into_data_bundle(args.image_path)
	save_as_fits(args.output_path, data_bundle)

if __name__=='__main__':
	exec_with_args(sys.argv[1:])


"""
Definitions of generic types so I can type hint more easily.

These may be unneeded and I should see if there is a better way to do what I want.
"""

from typing import TypeVar, GenericAlias, NewType

T = TypeVar('T')

type NTuple[T,N] = GenericAlias(tuple[T], N)

type NumVar = int | float

type ShapeVar[N] = NTuple[int,N]




"""
Module containing routines that operate on point spread function data
"""
from typing import Callable, TypeVar, Generic, ParamSpec, TypeVarTuple, Any
import functools
import itertools
import numpy as np
import scipy as sp
import scipy.ndimage

import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.array
import aopp_deconv_tool.numpy_helper.slice
from aopp_deconv_tool.optimise_compat import PriorParamSet

from aopp_deconv_tool.stats.empirical import EmpiricalDistribution

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


#DEBUGGING
import matplotlib.pyplot as plt


IntVar = TypeVar('IntVar', bound=int)
T = TypeVar('T')
Ts = TypeVarTuple('Ts')
P = ParamSpec('P')
Q = ParamSpec('Q')
class S(Generic[IntVar]): pass
class S1(Generic[IntVar]): pass
N = TypeVar('N',bound=int)
M = TypeVar('N',bound=int)


def normalise(
		data : np.ndarray, 
		axes : tuple[int,...] | None=None, 
		cutout_shape : tuple[int,...] | None = None,
		recentre_around_centre_of_mass = False,
		remove_background = True
	) -> np.ndarray:
	"""
	Ensure an array of data fufils the following conditions:
	
	* odd shape, to ensure a centre pixel exists
	* centre array on brightest pixel
	* ensure array sums to 1
	* cut out a region around the centre to remove unneeded data.
	"""
	if axes is None:
		axes = tuple(range(data.ndim))
	
	data[np.isinf(data)] = np.nan # ignore infinities
	data = nph.array.ensure_odd_shape(data, axes)
	
	
	# centre around brightest pixel
	for idx in nph.slice.iter_indices(data, group=axes):
		bp_offset = nph.array.get_centre_offset_brightest_pixel(data[idx])
		data[idx] = nph.array.apply_offset(data[idx], bp_offset)
		data[idx] /= np.nansum(data[idx]) # normalise
	
	if remove_background:
		# assume that the bottom-left corner is all background
		bg_region_slice = tuple(slice(0,data.shape[a]//10) for a in axes)
		for idx in nph.slice.iter_indices(data, group=axes):
			data[idx] -= np.median(data[idx][*bg_region_slice])
			data[idx] /= np.nansum(data[idx]) # normalise
	
	
	# cutout region around the centre of the image if desired,
	# this is pretty important when adjusting for centre of mass, as long
	# as the COM should be close to the brightest pixel
	if cutout_shape is not None:
		_lgr.debug(f'{tuple(data.shape[x] for x in axes)=} {cutout_shape=}')
		centre_slices = nph.slice.around_centre(tuple(data.shape[x] for x in axes), cutout_shape)
		_lgr.debug(f'{centre_slices=}')
		slices = [slice(None) for s in data.shape]
		for i, centre_slice in zip(axes, centre_slices):
			slices[i] = centre_slice
		_lgr.debug(f'{slices=}')
		data = data[tuple(slices)]
	
	
	if recentre_around_centre_of_mass:
		# move centre of mass to middle of image
		# threshold
		threshold = 1E-2
		with nph.axes.to_start(data, axes) as (gdata, gaxes):
			t_mask = (gdata > threshold*np.nanmax(gdata, axis=gaxes))
			_lgr.debug(f'{t_mask.shape=}')
			indices = np.indices(gdata.shape)
			_lgr.debug(f'{indices.shape=}')
			com_idxs = (np.nansum(indices*gdata*t_mask, axis=tuple(a+1 for a in gaxes))/np.nansum(gdata*t_mask, axis=gaxes))[:len(gaxes)].T
			_lgr.debug(f'{com_idxs.shape=}')
		
		_lgr.debug(f'{data.shape=}')
		
		for _i, (idx, gdata) in enumerate(nph.axes.iter_axes_group(data, axes)):
			_lgr.debug(f'{_i=}')
			_lgr.debug(f'{idx=}')
			_lgr.debug(f'{gdata[idx].shape=}')
			
			
			# calculate centre of mass
			#com_idxs = tuple(np.nansum(data[idx]*indices)/np.nansum(data[idx]) for indices in np.indices(data[idx].shape))
			centre_to_com_offset = np.array([com_i - s/2 for s, com_i in zip(gdata[idx].shape, com_idxs[idx][::-1])])
			_lgr.debug(f'{idx=} {com_idxs[idx]=} {centre_to_com_offset=}')
			_lgr.debug(f'{sp.ndimage.centre_of_mass(np.nan_to_num(gdata[idx]*(gdata[idx] > threshold*np.nanmax(gdata[idx]))))=}')
			
			# regrid so that centre of mass lies on an exact pixel
			old_points = tuple(np.linspace(0,s-1,s) for s in gdata[idx].shape)
			interp = sp.interpolate.RegularGridInterpolator(
				old_points, 
				gdata[idx], 
				method='linear', 
				bounds_error=False, 
				fill_value=0
			)
		
			# have to reverse centre_to_com_offset here
			new_points = tuple(p-centre_to_com_offset[i] for i,p in enumerate(old_points))
			_lgr.debug(f'{[s.size for s in new_points]=}')
			new_points = np.array(np.meshgrid(*new_points)).T
			_lgr.debug(f'{[s.size for s in old_points]=} {gdata[idx].shape=} {new_points.shape=}')
			gdata[idx] = interp(new_points)
	
	# Normalise again
	#for idx in nph.slice.iter_indices(data, group=axes):
	#	data[idx] /= np.nansum(data[idx])
	
	return data


def remove_offset(data, axes, mask, model_name='norm'):
	"""
	Remove the offsets of `data` over the specified `axes`, ignore any `mask`ed pixels.
	"""
	is_bad = np.isnan(data) | np.isinf(data)
	_lgr.debug(f'{is_bad.shape=}')
	
	
	match model_name:
		case 'norm':
			model = sp.stats.norm
		case 'gennorm':
			model = sp.stats.gennorm
		case _:
			raise RuntimeError(f'Unknown {model_name=} when fitting background for offset removal')
	
	param_names = ([x.strip() for x in model.shapes.split(',')] if model.shapes is not None else []) + ['loc', 'scale']
	
	
	nm_j_to_i = []
	noise_model_offsets = []
	noise_model_parameters = []
	noise_model_cdf = []
	noise_model_cdf_residual = []
	
	for i, (idx, gdata) in enumerate(nph.axes.iter_axes_group(data, axes)):
		if i % 10 == 0:
			_lgr.info(f'{i=}')
		_lgr.debug(f'{gdata[idx].shape=}')
		_lgr.debug(f'{mask[idx].shape=}')
		good_data = gdata[idx][mask[idx] & ~is_bad[idx]]
		_lgr.debug(f'{good_data.shape=}')
		
		if np.all(np.isnan(good_data)):
			_lgr.info(f'Skipping {i}^th offset calculation as whole slice is NANs')
			noise_model_offsets.append(np.nan)
			noise_model_parameters.append(dict([(k,np.nan) for k in param_names]))
			
			continue
			
		gd_distrib = EmpiricalDistribution(good_data.flatten())
		v, cdf = gd_distrib.whole_cdf()
		_lgr.debug(f'{v.shape=} {cdf.shape=}')
		
		
		params = model.fit(good_data)
		_lgr.debug(f'{param_names=} {params=}')
		noise_model_parameters.append(dict(zip(param_names, params)))
		_lgr.debug(f'{noise_model_parameters[-1]=}')
		noise_model_offsets.append(noise_model_parameters[-1]['loc'])
		
		good_data -= noise_model_offsets[-1]
		#gd_distrib = EmpiricalDistribution(good_data.flatten())
		#v, cdf = gd_distrib.whole_cdf()
		params = model.fit(good_data)
		
		noise_model_parameters[-1].update(dict(zip(param_names, params)))
	
	noise_model_at_values = []#np.linspace(np.nanmin(data), np.nanmax(data), 500)
		
	for i, (idx, gdata) in enumerate(nph.axes.iter_axes_group(data, axes)):
		if i % 100 == 0:
			_lgr.info(f'{i=}')
		good_data = gdata[idx][mask[idx] & ~is_bad[idx]]
		if np.all(np.isnan(good_data)):
			noise_model_at_values.append(np.full((500,), fill_value=np.nan))
			noise_model_cdf.append(np.full((500,), fill_value=np.nan))
			noise_model_cdf_residual.append(np.full((500,), fill_value=np.nan))
			continue
		gd_distrib = EmpiricalDistribution(good_data.flatten())
		
		noise_model_at_values.append(np.linspace(np.nanmin(good_data), np.nanmax(good_data), 500))
		noise_model_cdf.append(model(*noise_model_parameters[i].values()).cdf(noise_model_at_values[i]))
		noise_model_cdf_residual.append(noise_model_cdf[i] - gd_distrib.cdf(noise_model_at_values[i]))	
	
	
	return noise_model_offsets, noise_model_parameters, noise_model_at_values, noise_model_cdf, noise_model_cdf_residual


def trim_around_centre(
		data : np.ndarray,
		axes : tuple[int,...],
		output_shape : tuple[int,...]
	) -> np.ndarray:
	"""
	Truncate a numpy array about it's center.
	
	# ARGUMENTS #
	
	data : np.ndarray
		Data to trim around centre pixel, will remove pixels that are greater than shape[i]/2 from the centre pixel.
	axes : tuple[int,...]
		Axes over which to operate
	output_shape : tuple[int,...]
		Desired shape of `axes` after trimming
	"""
	_lgr.debug(f'{data.shape=}')
	_lgr.debug(f'{axes=}')
	_lgr.debug(f'{output_shape=}')
	old_shape = data.shape
	new_shape = list(old_shape)
	for i in range(data.ndim):
		if i in axes:
			j = axes.index(i)
			new_shape[i] = output_shape[j]
		else:
			new_shape[i] = old_shape[i]
	new_shape = tuple(new_shape)
	
	
	slices = [slice(None)]*data.ndim
	for i in range(data.ndim):
		if i in axes:
			centre_idx = data.shape[i]//2
			lo_idx = centre_idx - new_shape[i]//2
			hi_idx = centre_idx + new_shape[i]//2 + 1
			slices[i] = slice(lo_idx,hi_idx)
		else:
			slices[i] = slice(None)
	
	slices = tuple(slices)
	_lgr.debug(f'{slices=}')
	
	
	return data[slices]
	


def get_outlier_mask(
		data : np.ndarray,
		axes : tuple[int,...],
		n_sigma : float = 5,
	) -> np.ndarray:
	"""
	Get a boolean map of outliers in a numpy array. Assumes a normal distribution.
	
	# ARGUMENTS #
	
	data : np.ndarray
		Data to get outlier mask of.
	axes : tuple[int,...]
		Axes over which to get outlier mask
	n_sigma : float =  5
		Number of standard deviations away from the mean a member of `data` must be to be considered an outlier
	"""
	
	outlier_mask = np.zeros_like(data, dtype=bool)
	
	for _i, (idx, gdata) in enumerate(nph.axes.iter_axes_group(data, axes)):
		mean = np.nanmean(gdata[idx])
		std = np.nanstd(gdata[idx])
		
		outlier_mask[idx] = np.fabs(gdata[idx] - mean) > n_sigma*std
		
		label_map, n_labels = sp.ndimage.label(outlier_mask[idx])
		# order labels by number of pixels they contain
						
		ordered_labels = list(range(1,n_labels+1)) # label 0 is background
		ordered_labels.sort(key = lambda x: np.count_nonzero(label_map == x), reverse=True)
		
		# Only keep the largest region as we expect everything to have one (and only one) calibration source.
		outlier_mask[idx] *= False
		for i in range(1,n_labels):
			outlier_mask[idx] |= (label_map == ordered_labels[i])
		
		
	return outlier_mask


def get_roi_mask(
		data : np.ndarray, 
		axes : tuple[int,...], 
		threshold : float = 1E-2,
		n_largest_regions : None | int = 1,
	) -> np.ndarray:
	"""
	Find regions of interest in a numpy array by thresholding the array.
	
	## ARGUMENTS ##
	
	data : np.ndarray
		Array to get region of interest of
	axes : tuple[int,...]
		Axes to get region of interest along.
	threshold : float = 1E-2
		When finding region of interest, only values larger than this fraction of the maximum value are included.
	n_largest_regions : None | int = 1
		When finding region of interest, if using a threshold will only use the n_largest_regions in the calculation.
		A region is defined as a contiguous area where value >= `threshold` along `axes`. I.e., in a 3D cube, if
		we recentre about the COM on the sky (CELESTIAL) axes the regions will be calculated on the sky, not in
		the spectral axis (for example).
	"""
	mask = np.zeros_like(data, dtype=bool)
	with nph.axes.to_start(data, axes) as (gdata, gaxes), nph.axes.to_start(mask, axes) as (t_mask, t_axes):
	#with nph.axes.to_end(data, axes) as (gdata, gaxes):
		t_mask[...] = (gdata > threshold*np.nanmax(gdata, axis=gaxes))
		
		if n_largest_regions is not None:
			for j, (idx, g_mask) in enumerate(nph.axes.iter_axes_group(t_mask, gaxes)):
				label_map, n_labels = sp.ndimage.label(g_mask[idx])
				# order labels by number of pixels they contain
								
				ordered_labels = list(range(1,n_labels+1)) # label 0 is background
				ordered_labels.sort(key = lambda x: np.count_nonzero(label_map == x), reverse=True)
				
				if n_labels > n_largest_regions:
					g_mask[idx] *= False
					for i in range(n_largest_regions):
						g_mask[idx] |= (label_map == ordered_labels[i])
				
	return mask

def get_centre_of_mass_offsets(
	data : np.ndarray, 
		axes : tuple[int,...], 
		roi_mask : np.ndarray | None = None,
	) -> np.ndarray:
	"""
	Gets the location of the center of mass along `axes` (not an offset from the `data` center)
	
	## ARGUMENTS ##
	
	data : np.ndarray
		Array to recentre
	axes : tuple[int,...]
		Axes to get centre of mass and recentre along.
	roi_mask : np.ndarray | None = None
		Mask for the region of interest. If present will restrict calculations to this region.
	"""
	not_axes = tuple(i for i in range(data.ndim) if i not in axes)
	indices = np.indices(data.shape)
	com_offsets = np.moveaxis((np.nansum(indices*data[None,...]*roi_mask[None,...], axis=tuple(a+1 for a in axes))/np.nansum(data*roi_mask, axis=axes))[axes,...], 0, -1)
	
	return com_offsets

def get_brightest_pixel_offsets(
	data : np.ndarray, 
		axes : tuple[int,...], 
		roi_mask : np.ndarray | None = None,
	) -> np.ndarray:
	"""
	Gets the location of the brightest pixel (not an offset from the center)
	
	## ARGUMENTS ##
	
	data : np.ndarray
		Array to recentre
	axes : tuple[int,...]
		Axes to get centre of mass and recentre along.
	roi_mask : np.ndarray | None = None
		Mask for the region of interest. If present will restrict calculations to this region.
	"""
	import matplotlib.pyplot as plt # DEBUGGING
	
	data = np.array(data, dtype=data.dtype) # copy data
	
	axes_shape = tuple(s for i,s in enumerate(data.shape) if i in axes)
	not_axes = tuple(i for i in range(data.ndim) if i not in axes)
	indices = np.indices(data.shape) # shape = (n,s1,s2,s2,...,sn)
	
	data[~roi_mask] = np.nan
	offsets = np.zeros((len(axes), *tuple(s for i,s in enumerate(data.shape) if i not in axes)))
	for not_idxs in itertools.product(*tuple(range(data.shape[a]) for a in not_axes)):
	
		not_idx_iter = iter(not_idxs)
		not_slices = tuple(next(not_idx_iter) if a in not_axes else slice(None) for a in range(data.ndim))
		
		offsets[:,*not_idxs] = tuple(x for x,s in zip(np.unravel_index(np.nanargmax(data[not_slices]), shape=axes_shape), axes_shape))
		
	return np.moveaxis(offsets, 0, -1)

def apply_offsets(
		data : np.ndarray, 
		axes : tuple[int,...], 
		offsets : np.ndarray
	) -> np.ndarray:
	"""
	At the moment doesn't actually apply an offset, just shifts the data so the point at `offsets` is now the center.
	
	# Arguments #
	
	data : np.ndarray
		Array to recentre
	axes : tuple[int,...]
		Axes to get centre of mass and recentre along.
	offsets : np.ndarray
		Offsets to apply to data, will shift data's grid by this amount.
		
	"""
	_lgr.debug(f'{data.shape=}')
	_lgr.debug(f'{axes=}')
	_lgr.debug(f'{offsets.shape=}')
	
	for _i, (idx, gdata) in enumerate(nph.axes.iter_axes_group(data, axes)):
		_lgr.debug(f'{_i=}')
		_lgr.debug(f'{idx=}')
		_lgr.debug(f'{gdata[idx].shape=}')
		
		
		# calculate centre of mass
		#com_idxs = tuple(np.nansum(data[idx]*indices)/np.nansum(data[idx]) for indices in np.indices(data[idx].shape))
		#centre_to_com_offset = np.array([com_i - s/2 for s, com_i in zip(gdata[idx].shape, com_idxs[idx][::-1])])
		centre_to_com_offset = np.array([s//2 - com_i for s, com_i in zip(gdata[idx].shape, offsets[idx])])
		_lgr.debug(f'{idx=} {offsets[idx]=} {centre_to_com_offset=}')
		
		# regrid so that centre of mass lies on an exact pixel
		old_points = tuple(np.linspace(0,s-1,s) for s in gdata[idx].shape)
		interp = sp.interpolate.RegularGridInterpolator(
			old_points, 
			gdata[idx], 
			method='linear', 
			bounds_error=False, 
			fill_value=0
		)
	
		# have to reverse centre_to_com_offset here
		new_points = tuple(p-centre_to_com_offset[i] for i,p in enumerate(old_points))
		_lgr.debug(f'{[s.size for s in new_points]=}')
		new_points = np.array(np.meshgrid(*new_points)).T
		_lgr.debug(f'{[s.size for s in old_points]=} {gdata[idx].shape=} {new_points.shape=}')
		gdata[idx] = interp(new_points)
	return data



def objective_function_factory(model_flattened_callable, data, err, mode='minimise'):
	"""
	Given a model function, some data, and the error on that data; returns an objective function that
	for either 'minimise'-ing or 'maximise'-ing the difference/similarity of the model and data.
	"""
	match mode:
		case 'minimise':
			def model_badness_of_fit_callable(*args, **kwargs):
				residual = model_flattened_callable(*args, **kwargs) - data
				result = np.nansum((residual/err)**2)
				return np.log(result)
			return model_badness_of_fit_callable
		
		case 'maximise':
			def model_likelihood_callable(*args, **kwargs):
				residual = model_flattened_callable(*args, **kwargs) - data
				result = np.nansum((residual/err)**2)
				#_lgr.debug(f'{-np.log(result)=}')
				return -np.log(result)
			
			return model_likelihood_callable
		
		case _:
			raise NotImplementedError
	return

def scipy_fitting_function_factory(scipy_func):
	"""
	Given some function that implements the same protocol as [scipy minimise](https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize)
	returns a callable that accepts a PriorParamSet, a scipy-compatible objective function, a list of variable parameters, a list of constant parameters, and
	returns the fitted variable parameters.
	"""
	def scipy_fitting_function(params, objective_function, var_param_name_order, const_param_name_order):
		result = scipy_func(
			objective_function,
			tuple(params[p_name].const_value for p_name in var_param_name_order),
			bounds = tuple(params[p_name].domain for p_name in var_param_name_order)
		)
		return result.x

	return scipy_fitting_function

def fit_to_data(
		params : PriorParamSet,
		flattened_psf_model_callable : Callable[P, np.ndarray[S[N],T]],
		psf_data : np.ndarray[S[N],T],
		psf_err : np.ndarray[S[N],T],
		fitting_function : Callable[[PriorParamSet, Callable[Q,float]], Q],
		objective_function_factory : Callable[[Callable[P,np.ndarray[S[N],T]], np.ndarray[S[N],T], np.ndarray[S[N],T]], Callable[Q,float]] = functools.partial(objective_function_factory, mode='minimise'),
		plot_mode : str | bool | None = None
	) -> tuple[np.ndarray[S[N],T], dict[str,Any], dict[str,Any]]:
	"""
	Fits a model to some data with some error on that data.
	"""
	
	model_scipyCompat_callable, var_param_name_order, const_param_name_order = params.wrap_callable_for_scipy_parameter_order(
		flattened_psf_model_callable, 
		arg_names=flattened_psf_model_callable.arg_names if hasattr(flattened_psf_model_callable,'arg_names') else None
	)
	
	# Possibly take this plotting code out
	if plot_mode is not None:
		import matplotlib.pyplot as plt
		plt.close()
		test_result = model_scipyCompat_callable(tuple(params[p_name].const_value for p_name in var_param_name_order))
		
		f, ax = plt.subplots(2,2,squeeze=False,figsize=(12,8))
		ax = ax.flatten()
		
		f.suptitle(f'Example Plot\nvariables {dict((p.name,p.const_value) for p in params.variable_params)}\n constants {dict((p.name,p.const_value) for p in params.constant_params)}')
		
		ax[0].imshow(test_result)
		ax[0].set_title('example psf')
		
		ax[1].plot( test_result[:,test_result.shape[1]//2], np.arange(test_result.shape[0]),)
		ax[1].set_xscale('log')
		ax[1].set_title('y marginalisation')
		
		ax[2].plot(np.arange(test_result.shape[1]), test_result[test_result.shape[0]//2,:])
		ax[2].set_yscale('log')
		ax[2].set_title('x marginalisation')
		
		ax[3].remove()
		
		plt.show()
	
	
	objective_function = objective_function_factory(model_scipyCompat_callable, psf_data, psf_err)
	
	
	fitted_params = fitting_function(params, objective_function, var_param_name_order, const_param_name_order)
	
	return (
		model_scipyCompat_callable(fitted_params), 
		dict((k,v) for k,v in zip(var_param_name_order, fitted_params)), 
		dict((p.name,p.const_value) for p in params.constant_params)
	)


from typing import Any
from types import TracebackType

class Alias:
	"""
	Context manager wrapper around a variable to temporarily change it's name
	"""
	slots=('ptr',)
	def __init__(self, var : Any):
		self.ptr = var
	def __enter__(self):
		return(self.ptr.val)
	def __exit__(self, etype : type, evalue : Exception, traceback : TracebackType):
		return(False)


"""
Classes and functions for working with command-line arguments alongside the `argparse` module.
"""

import sys, os
import dataclasses as dc
import re
import typing
from typing import Any, Type
import argparse

import aopp_deconv_tool
import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')

re_empty_line = re.compile(r'^\s*$\s*', flags=re.MULTILINE)

class DataclassArgFormatter (argparse.RawTextHelpFormatter):#, argparse.MetavarTypeHelpFormatter):
	"""
	Formatter for command-line arguments found from a dataclass.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

class TypeConverterFactory:
	"""
	When instantiated with a type, returns a callable that converts a passed string to that type.
	"""
	def __init__(self, type : Type):
		"""
		Register which type the converter will convert strings to.
		"""
		self.type = type
		self.meta_type = typing.get_origin(type)
		self.types = typing.get_args(type)
	
	def __call__(self, astring : str):
		"""
		Convert `astring` to the type specified on object creation.
		"""
		if self.meta_type is None:
			return self.type(astring)
		else:
			for atype in self.types:
				return atype(astring)
		raise RuntimeError(f"Cannot convert '{astring}' to any of the supported types for {self.type}")

def parse_args_of_dataclass(
		dataclass     : Type, 
		argv          : list[str] | tuple[str], 
		show_help     : bool                   = False, 
		prog          : str                    = None, 
		description   : str                    = None, 
		arg_prefix    : str                    = '', 
		metadata_keys : list[str]              = None,
	) -> dict[str,Any]:
	"""
	Use the `argparse` package to read the fields of a DataClass. Fields that are "init" fields will be
	used as command-line arguments. A "description" entry will be looked for in a field's `metadata` dict,
	and used if it is found. Type conversion is handled by converting command-line argument string to the
	field's type as specified in the `field.type` attribute. The type conversion uses `TypeConverterFactory`.
	"""
	
	parser = argparse.ArgumentParser(
		prog=prog,
		description = re_empty_line.split(dataclass.__doc__,2)[1] if description is None else description, 
		formatter_class=DataclassArgFormatter,
		add_help=False
	)
	
	def on_parser_error(err_str):
		print(err_str)
		parser.print_help()
		sys.exit(1)
	
	parser.error = on_parser_error
	
	max_string_length = 50
	try:
		max_string_length = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	# always want a 'defaut' value, but will get it from the field, but add it here so
	# padding is correct
	if metadata_keys is None:
		metadata_keys = {'default'}
		for field in dc.fields(dataclass):
			metadata_keys.update(set(field.metadata.keys()))
		metadata_keys.discard('description') # discard this as the description does not get printed with it's key.
	else:
		metadata_keys.update({'default'})
	
	max_metadata_key_size = max(len(k) for k in metadata_keys)
	metadata_keys.discard('default') # We don't actually have a "default" entry, so discard it
	
	# Get correct format string so colons line up when metadata is printed
	metadata_fmt = '{:<'+str(max_metadata_key_size)+'} : {}'
	
	negative_bool_flag_args = []
	for field in dc.fields(dataclass):
		if field.init != True: # only include parameters that are passed to init
			continue
			
		field_default = field.default if field.default != dc.MISSING else (field.default_factory() if field.default_factory != dc.MISSING else None)
		
		field_arg_string = f'--{arg_prefix}{field.name}'
		field_help_string = aopp_deconv_tool.text.wrap(
			'\n'.join((
				field.metadata.get('description', 'DESCRIPTION NOT FOUND'),
				metadata_fmt.format('default', str(field_default)),
				*(metadata_fmt.format(k,field.metadata.get(k, f'{k.upper()} NOT FOUND')) for k in metadata_keys)
			)),
			max_string_length,
			combine_strings_of_same_indent_level = False
		)
		
		field_type_string = str(field.type)
		if field_type_string.startswith("<class '") and field_type_string.endswith("'>"):
			field_type_string = field_type_string[8:-2]
		elif field_type_string.startswith("typing.Optional[") and field_type_string.endswith(']'):
			field_type_string = 'None | ' + field_type_string[16:-1]
		else:
			field_type_string = 'UNKNOWN TYPE'
		
		
		if field.type is bool:
			arg_string = field.name
			if field_default is True:
				arg_string = 'no_'+field.name
				negative_bool_flag_args.append(arg_string)
				
			parser.add_argument(
				'--'+arg_string,
				#type=field.type, 
				action = 'store_true' if field_default is False else 'store_false',
				help= field_help_string,
				#metavar=str(field.type)[8:-2]
			)
		else:
			_lgr.debug(f'{field.name=} {field.type=}')
			parser.add_argument(
				'--'+field.name, 
				type=TypeConverterFactory(field.type), 
				default= field_default,
				help= field_help_string,
				metavar=field_type_string
			)
	
	
	args = vars(parser.parse_args(argv))
	
	if show_help:
		parser.print_help()
		sys.exit()
	
	# Remove the "no_" prefix from negative boolean flags
	for k in negative_bool_flag_args:
		args[k[3:]] = args[k]
		del args[k]
	
	return args


def construct_arglist_from_locals(
		dict_of_locals : dict[str, Any],
		n_positional_args : int = 0
	):
	"""
	Given a key-value dictionary, construct something we can pass to an argparse parser.
	Entries with `None` as their value will NOT be constructed into the list.
	"""
	dict_of_locals.update(dict_of_locals.get('kwargs',{}))
	if 'kwargs' in dict_of_locals:
		del dict_of_locals['kwargs']
	
	arglist = []
	
	for i,(k,v) in enumerate(dict_of_locals.items()):
		# Positional arguments get stuffed in without names
		if i < n_positional_args:
			arglist.append(str(v))
			continue
		# `None` means that we didn't pass anything, so don't send to the argument parser.
		if v is not None:
			# if the key ends in "_FLAG" we don't send a value, and remove the suffix
			if k.endswith('_FLAG'):
				arglist.append(f'--{k[:-5]}')
			else:
				arglist.append(f'--{k}')
				if type(v) in (list, tuple):
					arglist.extend([str(x) for x in v])
				else:
					arglist.append(str(v))
	
	_lgr.debug(f'{arglist=}')
	return arglist





import numpy as np

from aopp_deconv_tool.instrument_model.instrument_base import InstrumentBase
from aopp_deconv_tool.geometry.shape import Circle
from aopp_deconv_tool.optics.geometric.optical_component import OpticalComponentSet, Aperture, Obstruction, Refractor
from aopp_deconv_tool.optics.telescope_model import optical_transfer_function_of_optical_component_set

class VLT(InstrumentBase):
	"""
	Instrument description for very large telescope.
	"""
	n_actuators = 28#42 # fiddle with this, somewhere between 24 and 42 seems likely
	obj_diameter = 8 # meters
	f_ao = n_actuators / (2*obj_diameter)
	primary_mirror_focal_length = 120 # meters
	primary_mirror_pos = 100 # meters
	primary_mirror_diameter = 8
	secondary_mirror_diameter_frac_of_primary = 2.52/18
	secondary_mirror_dist_from_primary =50 # meters
	secondary_mirror_diameter_meters = primary_mirror_diameter*secondary_mirror_diameter_frac_of_primary
	ocs = OpticalComponentSet.from_components([
		Aperture(
			0, 
			'objective aperture', 
			shape=Circle.of_radius(obj_diameter/2)
		), 
		#Obstruction(
		#	primary_mirror_pos - secondary_mirror_dist_from_primary, 
		#	'secondary mirror back', 
		#	shape=Circle.of_radius(secondary_mirror_diameter_meters/2)
		#), 
		Refractor(
			primary_mirror_pos, 
			'primary mirror', 
			shape=Circle.of_radius(primary_mirror_diameter/2), 
			focal_length=primary_mirror_focal_length
		),
	])
	
	
	@classmethod
	def muse(cls,
			expansion_factor, # Increase this to get better resolution
			supersample_factor, # Increase this to get better area coverage
			obs_shape = (201,201),
			obs_pixel_size = 0.025 / (3600) *np.pi/180,
			ref_wavelength = 5E-7
		):
		"""
		Description of MUSE instrument on the VLT telescope
		"""
		obs_scale = np.array(tuple(s/cls.obj_diameter for s in obs_shape))
		#obs_scale = np.array(obs_shape)*np.array(obs_pixel_size)/ref_wavelength
		return cls(obs_shape, obs_scale, obs_pixel_size, ref_wavelength, expansion_factor, supersample_factor)





"""
Text wrapping, filling, formatting, indenting etc.
"""

import dataclasses as dc
from collections import namedtuple
from typing import Iterable
import textwrap
import re

import aopp_deconv_tool.cast as cast

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')


newline='\n'
tab='\t'
space=' '

def indent(x : str, n : int, s : str = tab):
	"""
	Indent all lines in `x` by `n` copys of string `s`
	"""
	return textwrap.indent(x, n*s, lambda _: True)

def to_tuple(x : str, t : type[type] | Iterable[type]):
	"""
	Converts a text representation of a tuple to a tuple of the specified type.
	
	Arguments:
		x : str
			string to convert to a tuple
		t : type[type] | Iterable[type]
			Type of tuple to convert to. If `t` is not an iterable, then will
			convert to a homogeneous tuple, otherwise will convert each element
			of `x` to the corresponding element of `t`
	"""
	assert type(x) == str
	
	# if x ends with a comma, remove it
	if x.endswith(','):
		x = x[:-1]
	
	if isinstance(t, type):
		# assume tuple homogeneous
		try:
			return tuple(cast.to(y,t) for y in x.split(','))
		except Exception as e:
			e.add_note(f"string '{x}' could not be cast to a homogeneous {t} tuple")
			raise
	
	assert len(t) == x.count(',')+1, "Must have same number of tuple entries as types"
	try:
		return tuple(cast.to(y,u) for y,u in zip(x.split(','), t))
	except Exception as e:
		e.add_note(f"string '{x}' could not be cast to a tuple of types {t}")
		raise
	
@dc.dataclass
class BracketState:
	"""
	Class for holding the depth of nested brackets when parsing text.
	"""
	r : int = 0 # round
	c : int = 0 # curly
	s : int = 0 # square
	
	def update(self, c : str):
		"""
		Given a character `c` will update the nested bracket depth counters
		"""
		if c == '[': self.s+=1; return
		if c == ']': self.s-=1; return
		if c == '{': self.c+=1; return
		if c == '}': self.c-=1; return
		if c == '(': self.r+=1; return
		if c == ')': self.r-=1; return
	
	def total(self) -> int:
		"""
		Returns the sum of the bracket counters.
		"""
		return self.r + self.c + self.s


def split_around_brackets(x : str, sep : str = ','):
	"""
	Split on a separator `sep`, do not split on any separator inside any types
	of bracket.
	"""
	bstate = BracketState(0,0,0)
	stack = []
	idx = 0
	for i, c in enumerate(x):
		_lgr.debug(i,c)
		if c == sep and bstate.total()==0: 
			stack.append(x[idx:i])
			idx = i+1
			continue
		bstate.update(c)
	stack.append(x[idx:])
	return(stack)



initial_whitespace = re.compile(r'^\s*')
str_break_chars = re.compile(r'\s')

def combine_lines_with_same_indent(x : str, preserve_repeated_empty_lines=False):
	y = []
	j=0
	
	current_block = []
	i_span_current = -1
	for i,z in enumerate(x.split(newline)):
		iw_match = initial_whitespace.match(z)
		i_span = iw_match.span()[1] - iw_match.span()[0]
		
		_lgr.debug(f'{z=}')
				
		# If line is empty, start a new block if we preserve empty lines else do not.
		if z.isspace() or len(z)==0:
			i_span = -1 if not preserve_repeated_empty_lines else (-1 if i_span_current > 0 else (i_span_current-1))
		
		_lgr.debug(f'{i_span=} {i_span_current=}')
		
		# if we have changed indent span, we move on to the next block
		if i_span != i_span_current:
			y.append(space.join(current_block))
			current_block = []
			current_block.append(z.rstrip())
			i_span_current = i_span
		else:
			current_block.append(z[i_span:].rstrip())
		
		_lgr.debug(f'{i_span=} {i_span_current=}')
		_lgr.debug(f'{current_block=}')
			
	return newline.join(y)



def wrap(x : str, width=70, combine_strings_of_same_indent_level=True):
	x = combine_lines_with_same_indent(x, preserve_repeated_empty_lines=False) if combine_strings_of_same_indent_level else x

	x.replace(tab, '    ')
	y = x.split(newline)

	for i, z in enumerate(y):
		if len(y[i]) > width:
			iw_match = initial_whitespace.match(y[i])
			iw = iw_match.group() if iw_match is not None else ''
			b_search = str_break_chars.search(y[i][len(iw):width][::-1])
			b_idx = width - b_search.span()[0] if b_search is not None else width
			y.insert(i+1, iw+y[i][b_idx:])
			y[i] = y[i][:b_idx]
	return newline.join(y)
